

/**
 * Provides an interface for classes that allow sentence pairs and
 * word lists to be looked up from a corpus. In the context of this
 * class and those that implement it, source and dest refer to
 * the source and destination languages of the translation
 * system being trained from this corpus.
 */

public interface CorpusAccessor {


    /**
     * Returns the length of the source or destination language sentence
     * in one of the sentence pairs of the corpus.

     @param index the index of the sentence pair in question
     @param isDestLanguage whether the query is about the source-langauge
     * sentence in the pair or about the destination-language sentence

     @return the length of the the specified sentence
     */

    public int getSentenceLength(int index, boolean isDestLanguage);

    /**
     * Returns the word at the given index in the specified sentence
     * from the specified sentence pair of the corpus.

     @param sentenceIndex the index of the sentence pair in question
     @param isDestLanguage whether the query is about the source-langauge
     * sentence in the pair or about the destination-language sentence
     @param wordIndex the index of the word in the given sentence

     @return the specified word
     */

    public String getWord(int sentenceIndex, boolean isDestLanguage, int wordIndex);

    /**
     * Returns the source or destination language sentence from the
     * specified sentence pair.

     @param index the index of the sentence pair in question
     @param isDestLanguage whether the query is about the source-langauge
     * sentence in the pair or about the destination-language sentence

     @return the specified sentence, as an array of strings,
     * one word per string
     */
    
    public String[] getSentence(int index, boolean isDestLanguage);

    /**
     * Returns the source and destination language sentences of the
     * specified sentence pair.

     @param index the index of the sentence pair in question

     @return the specified sentence pair, as an array containing the
     * array representation of the source sentence and the array
     * representation of the second
     */
    
    public String[][] getSentencePair(int index);

    /**
     * Returns the number of sentence pairs in the corpus.

     @return the number of sentence pairs
     */
    
    public int getSize();

    /**
     * Returns an array of all the unique words in the specified language.
     * Implementations may return the same array object on subsequent
     * calls, so callers should not modify the result.

     @param isDestLanguage whether the list of unique words from the
     * source language or the destination language is desired

     @return the chosen word list
    */
    
    public String[] getWordList(boolean isDestLanguage);
    
}
