
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Hashtable;

// import java.io.Console;

/**
 * Implements the CorpusAccessor interface to provide access to corpus
 * data. Corpus data is loaded from a pair of text files, one containing
 * the source language sentences and the other the destination language
 * sentences, such that lines <code>n</code> in the source and destination
 * language files contain the two halves of a sentence pair. Words in
 * a sentence are space-separated. Within this class, source and dest
 * refer to the source and destination languages of the translation
 * system being trained from this corpus.
 */

public class CorpusFromTextPair implements CorpusAccessor {
	
	private String sourceFile, destFile;
	private ArrayList<String[]> sourceSentences, destSentences;
	private String[] sourceWords, destWords;
	private Hashtable<String, HashSet<String>> wordMap;

	/**
	 * Loads corpus data from the text files at the specified
	 * filenames.

	 @param sourceFile the name of the text file containing the
	 * source language data.
	 */
	
	public CorpusFromTextPair(String sourceFile, String destFile)
	{
		this.sourceFile = sourceFile;
		this.destFile = destFile;
		sourceSentences = new ArrayList<String[]>();
		destSentences = new ArrayList<String[]>();
		sourceWords = null;
		destWords = null;
		
		loadFromFiles();
	}

	/**
	 * Loads the corpus data from the filenames set during the
	 * constructor.
	 */
	
	public void loadFromFiles()
	{
		loadToArrayList(sourceFile, sourceSentences);
		loadToArrayList(destFile, destSentences);
		if (sourceSentences.size() != destSentences.size()) {
			System.out.println("Warning: Files are not the same length!");
		}
	}

	/**
	 * Formats and prints an ArrayList of arrays of Strings for
	 * debugging purposes.

	 @param list the ArrayList to be printed
	 */
	
	public String arrayListToString(ArrayList<String[]> list)
	{
		String output = "[";
		for (String[] array : list) {
			output += "[";
			for (String s : array) {
				output += s + ", ";
			}
			output += "], ";
		}
		output += "]";
		return output;
	}

	/**
	 * Loads the specified file and inserts array-of-Strings
	 * representations of the sentences constituting each line of the
	 * file into the given ArrayList.

	 @param fileName the name of the file to load
	 @param sentenceList the ArrayList to write the loaded sentence
	 * arrays to
	 */
	
	public void loadToArrayList(String fileName, ArrayList<String[]> sentenceList)
	{
		BufferedReader inputStream = null;
		try {
			inputStream = new BufferedReader(new FileReader(fileName));
			String line;
			// System.out.println("Loading file " + fileName);
			while ((line = inputStream.readLine()) != null) {
				//do stuff
				String [] splitLine = line.split("\\s+");
				int emptyWordCount = 0;
				for(String w : splitLine)
				{
					if(w.length() == 0)
					{
						//System.out.println("FOUND length-0 word");
						emptyWordCount++;
					}
				}
				String [] trimmedSplitLine = new String[splitLine.length - emptyWordCount];
				int insertIndex = 0;
				for(int i = 0; i < splitLine.length; i++)
				{
					if(splitLine[i].length() > 0)
						trimmedSplitLine[insertIndex] = splitLine[i];
						insertIndex++;
				}
				//if(emptyWordCount > 0)
					//System.out.println(Arrays.toString(trimmedSplitLine));
				sentenceList.add(trimmedSplitLine);
			}
			//System.out.print(arrayListToString(sentenceList));
		} catch (IOException e) {
			System.out.println("Could not open " + fileName);
			sentenceList = new ArrayList<String[]>();
			return;
		}
		
	}

	/**
	 * Returns the length of the source or destination language sentence
	 * in one of the sentence pairs of the corpus.

	 @param index the index of the sentence pair in question
	 @param isDestLanguage whether the query is about the source-langauge
	 * sentence in the pair or about the destination-language sentence

	 @return the length of the the specified sentence
	 */

	public int getSentenceLength(int index, boolean isDestLanguage)
	{
		int length = getSentence(index, isDestLanguage).length;
		return length;
	}

    /**
     * Returns the word at the given index in the specified sentence
     * from the specified sentence pair of the corpus.

     @param sentenceIndex the index of the sentence pair in question
     @param isDestLanguage whether the query is about the source-langauge
     * sentence in the pair or about the destination-language sentence
     @param wordIndex the index of the word in the given sentence

     @return the specified word
     */
    
    public String getWord(int sentenceIndex, boolean isDestLanguage, int wordIndex)
    {
    	String word = getSentence(sentenceIndex, isDestLanguage)[wordIndex];
    	return word;
    }

    /**
     * Returns the source or destination language sentence from the
     * specified sentence pair.

     @param index the index of the sentence pair in question
     @param isDestLanguage whether the query is about the source-langauge
     * sentence in the pair or about the destination-language sentence

     @return the specified sentence, as an array of strings,
     * one word per string
     */
    
    public String[] getSentence(int index, boolean isDestLanguage)
    {
    	String[] sentence = (isDestLanguage ? destSentences : sourceSentences).get(index);
    	return sentence;
    }

    /**
     * Returns the source and destination language sentences of the
     * specified sentence pair.

     @param index the index of the sentence pair in question

     @return the specified sentence pair, as an array containing the
     * array representation of the source sentence and the array
     * representation of the second
     */
    
    public String[][] getSentencePair(int index)
    {
    	String[][] sentencePair = {getSentence(index, false), getSentence(index, true)};
    	return sentencePair;
    }

    /**
     * Returns the number of sentence pairs in the corpus.

     @return the number of sentence pairs
     */
    
    public int getSize()
    {
    	int length = sourceSentences.size();
    	return length;
    }

    /**
     * Returns an array of all the unique words in the specified language.
     * The first time the function is called, the array is generated
     * by iterating through all the sentence pairs in the corpus.
     * The array is cached so that subsequent calls immediately return
     * the cached array; thus, callers MUST NOT modify the array contents.

     @param isDestLanguage whether the list of unique words from the
     * source language or the destination language is desired

     @return the chosen word list
    */
    
    public String[] getWordList(boolean isDestLanguage)
    {
    	if(isDestLanguage && destWords != null)
    		return destWords;
    	if(!isDestLanguage && sourceWords != null)
    		return sourceWords;
    	//Lazy strategy
    	ArrayList<String[]> sentenceList = (isDestLanguage ? destSentences : sourceSentences);
    	ArrayList<String> wordList = new ArrayList<String>();
    	HashSet<String> wordSet = new HashSet<String>();
    	for(String[] sentence : sentenceList) {
    		for(String word : sentence) {
    			//if(!wordList.contains(word)) {
    			if(!wordSet.contains(word))
    			{
    				wordList.add(word);
    				wordSet.add(word);
    			}
    		}
    	}
    	
    	String[] temp = {};
    	String[] result = wordList.toArray(temp);
    	if(isDestLanguage)
    		destWords = result;
    	else
    		sourceWords = result;
    	return result;
    }
    
    public Hashtable<String, HashSet<String>> getWordMap() {
        if (wordMap != null) 
            return wordMap;
        wordMap = new Hashtable<String, HashSet<String>>();
        for (int i=0; i < getSize(); i++)
		{
		    String[] sourceSentence = getSentence(i, false);
		    String[] destSentence = getSentence(i, true);
		    //System.out.println("processing sentences: " + Arrays.toString(sourceSentence));
		    
		    for (String sourceWord : sourceSentence)
		    {
		        for (String destWord : destSentence) 
		        {
		            if (!wordMap.containsKey(sourceWord))
		                wordMap.put(sourceWord, new HashSet<String>());
		            wordMap.get(sourceWord).add(destWord);
		        }
		    }
		}
		assert(wordMap != null);
		return wordMap;
    }

    /**
     * Tests the corpus implementation on a small testing corpus
     * in the repository.

     @param args the command-line arguments, which are ignored
     */
    
    public static void main(String[] args)
    {
    	System.out.println("Testing accuracy of import...");

    	CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/spanish.txt", "../corpora/textpairs/english.txt");
    	assert(corpus.getSentenceLength(2, false) == 4);
    	assert(corpus.getSentenceLength(0, true) == 9);
    	assert(corpus.getWord(3, false, 1).equals("gato"));
    	assert(corpus.getWord(3, false, 1).equals("cat"));
    	assert(corpus.getSentence(4, false)[3].equals("rojo"));
    	assert(corpus.getSentence(4, true)[3].equals("red"));
    	assert(corpus.getSentencePair(4)[0][3].equals("rojo"));
    	assert(corpus.getSentencePair(4)[1][3].equals("red"));
    	assert(corpus.getSize() == 5);
    	assert(corpus.getWordList(false).length == 17);
    	assert(corpus.getWordList(true).length == 14);
    	System.out.println("Test successful");

    }
}
