import java.util.*;
import java.io.*;
import org.apache.commons.math.util.MathUtils;


/**
 * Implements the alternating search decoding algorithm.
 */
public class ASDecoder implements Decoder {

    private Model3Params params;
    private CorpusAccessor corpus;
    private LanguageModel languageModel;
    private Hashtable<String, TableEntry> currentRow;
    private Hashtable<String, TableEntry> prevRow;
    private HashSet<String> uniqueSourceWordSet;
    public String[] uniqueDestWords;
    private SortedProbabilityAccessor nextToPrev; // used to find likely 'previous words'
    private Hashtable<String, String[]> translationCandidates;
    private String[] topWords;

    // unknown word handling
    // this isn't currently working; currently we just remove all unknown words
    protected boolean hasUnknownWords;
    protected HashSet<String> unknownSourceWords; 
    protected HashSet<String> unknownSourceWordPairs;
    protected String[] unknownSourceWordsArray;
    protected WordsAndProbabilities knownToUnknownWAP;
    protected WordsAndProbabilities unknownToKnownWAP;
    protected static double unknownBigramProb = -6;
    
    /**
     * Constructs an ASDecoder object, for specific parameters

     @param p The parameters to use
     @param c The corpus accessor to use
     @param l Language model to use (gets bigram probabilities) 
     @param ntp Used to optimize the process of finding a previous word
    */
    public ASDecoder(Model3Params p, CorpusAccessor c, LanguageModel l, SortedProbabilityAccessor ntp) {
	params = p;
	corpus = c;
	languageModel = l;
	nextToPrev = ntp;

	String[] uniqueSourceWords = corpus.getWordList(false);
	uniqueDestWords = corpus.getWordList(true);
	setSortedDestWords();

	uniqueSourceWordSet = new HashSet<String>();

	int numTopWords = 200;
	
	topWords = new String [numTopWords];
	for(int i = 0; i < numTopWords; i++) {
		topWords[i] = uniqueDestWords[i];
	}

	for (String word : uniqueSourceWords) {
	    uniqueSourceWordSet.add(word);
	}
    }

    /**
     * Finds unknown words in the source sentence. Curently does nothing with them

     @param sourceSentence The source sentence to search for unknown words
    */
    private void findUnknowns(String[] sourceSentence) {
	unknownSourceWords = new HashSet<String>();
	unknownSourceWordPairs = new HashSet<String>();
	ArrayList<String> unknownSourceWordsList = new ArrayList<String>();
	hasUnknownWords = false;
	String prevUnknownWord = null;
	int index = 0;
	for (String word : sourceSentence) {
	    if (! uniqueSourceWordSet.contains(word)) {
		System.out.println("found unknown: "+word);
		hasUnknownWords = true;
		unknownSourceWords.add(word);
		unknownSourceWordsList.add(word);
		index++;
		prevUnknownWord = word;
		if (prevUnknownWord != null) {
		    unknownSourceWordPairs.add(prevUnknownWord + " " + word);
		}
	    } else {
		prevUnknownWord = null;
	    }
	}

	// set up WAP
	unknownSourceWordsArray = unknownSourceWordsList.toArray(new String[unknownSourceWordsList.size()]);
	double[] lowProbs = new double[unknownSourceWords.size()];
	for (int i = 0; i < lowProbs.length; i++) {lowProbs[i] = unknownBigramProb;}
	knownToUnknownWAP = new WordsAndProbabilities(unknownSourceWordsArray, lowProbs);
	
	lowProbs = new double[uniqueDestWords.length];
	for (int i = 0; i < lowProbs.length; i++) {lowProbs[i] = unknownBigramProb;}
	unknownToKnownWAP = new WordsAndProbabilities(uniqueDestWords, lowProbs);

	uniqueDestWords = Viterbi.concat(uniqueDestWords, unknownSourceWordsArray);
    }
    

    /**
     * Removes unknown words from the input sentence

     @param sourceSentence The string array to search for unknown words
     @return The new string array with unknown words removed
    */
    private String[] removeUnknownWords(String[] sourceSentence) {
	ArrayList<String> knownSourceWordsList = new ArrayList<String>();
	int index = 0;
	for (String word : sourceSentence) {
	    if (uniqueSourceWordSet.contains(word)) {
		knownSourceWordsList.add(word);
		index++;
	    }
	}
	return knownSourceWordsList.toArray(new String[knownSourceWordsList.size()]);
    }

    /**
     * Implements the high level AS algorithm. This includes calling a method
     * to find a new alignment (Model3Viterbi()), and a method to find a new sentence (fixedAlignmentDecode)

     @param sourceSentence The sentence to decode, in the source language
     @return The Decoded string, in the target language
    */
    public String decode(String sourceSentence) {
	String[] splitSourceSentence = sourceSentence.split("\\s+");
	int m = splitSourceSentence.length;

	getAllBestWords(splitSourceSentence); // finds word translation candidates for all words in the source sentence
	splitSourceSentence = removeUnknownWords(splitSourceSentence);
	findUnknowns(splitSourceSentence); // note that this will never find unknown words, as it is called after they're removed

	String[] bestE = null; // the current best destination language sentence (this is our final output)
	Alignment bestA = null; // the current best alignment
	double bestProb = Double.NEGATIVE_INFINITY;

	for (int l = m/2; l <= 1.5*m; l++) {
	    if (l==0) { l=1; }
	    Alignment a = null;
	    String[] e = null;
	    
	    double oldProb = Double.NEGATIVE_INFINITY;
	    double currentProb = -1000;
	    while (currentProb > oldProb) {
		e = naiveDecode(splitSourceSentence, l, a); // finds a new dest sentence

		Alignment model2best = Viterbi.viterbi(e, splitSourceSentence, params);
		a = Viterbi.model3ViterbiApprox(model2best, e, splitSourceSentence, params); // finds a new alignment
		oldProb = currentProb;
		currentProb = Viterbi.getAlignmentAndTranslationProb(a, splitSourceSentence, e, params);
	    }
	    if (bestE == null || bestA == null || oldProb > bestProb) {
		bestProb = oldProb;
		bestE = e;
		bestA = a;
	    }
	}
	String result = "";
	for (String s : bestE) {
	    result += s + " ";
	}
	return result.trim();
    }

    /**
     * Wrapper around fixedAlignmentDecode. Chooses an arbitrary alignment if none is given

     @param sourceSentence The sentence to decode
     @param length The target length to decode to
     @param a The alignment to use in decoding
     @return The decoded sentence
    */
    private String [] naiveDecode(String[] sourceSentence, int length, Alignment a) {
	if (a == null) {
	    int [] alignmentArray = new int[sourceSentence.length];
	    for (int i = 0; i < sourceSentence.length; i++) { alignmentArray[i] = i%length+1; }
	    a = new Alignment(alignmentArray);
	}
	return fixedAlignmentDecode(sourceSentence, length, a);
    }

    /**
     * Finds an approximately optimal destination language sentence given a source language
     * sentence and an alignment.

     @param sourceSentence The sentence to decode
     @param length The target length to decode to
     @param a The alignment to use in decoding
     @return The decoded sentence
    */
    private String [] fixedAlignmentDecode(String[] sourceSentence, int length, Alignment a) {
	ArrayList<Hashtable<String, TableEntry>> dpTable = new ArrayList<Hashtable<String, TableEntry>>(length);
	Hashtable<String, TableEntry> currentRow = new Hashtable<String, TableEntry>();
	String[] candidates = getAllCandidates(sourceSentence, a, 1);
	if (candidates.length == 0) { candidates = uniqueDestWords; }

	// base case for the DP table: partial sentences of length 1, i.e. 1-word sentence
	for (String word : candidates) {
	    WordsAndProbabilities [] prevWordsAndProbs = getPrevWordsAndProbs(word);
	    String[] prevWords = Viterbi.concat(prevWordsAndProbs[0].getWords(), prevWordsAndProbs[1].getWords());
	    double[] probs1 = prevWordsAndProbs[0].getProbs();
	    double[] probs2 = prevWordsAndProbs[1].getProbs();
	    double[] prevProbs = Arrays.copyOf(probs1, probs1.length + probs2.length);
	    System.arraycopy(probs2, 0, prevProbs, probs1.length, probs2.length);
	    for (int i = 0; i < prevWords.length; i++) {
		if (prevWords[i].equals("<s>")) {
		    currentRow.put(word, new TableEntry(word, getAlignedWordProb(word, 1, a, sourceSentence, length)+prevProbs[i], null));
		    break;
		}
	    }
	}
	dpTable.add(currentRow); // add the base case row to the table
	
	// general case for the table: partial sentences between length 2 and the target length
	for (int i = 2; i <= length; i++) {
	    currentRow = new Hashtable<String, TableEntry>();
	    prevRow = dpTable.get(i-2);
	    candidates = getAllCandidates(sourceSentence, a, i);
	    if (candidates.length == 0) { candidates = uniqueDestWords; }
	    for (String word : candidates) {
		WordsAndProbabilities [] prevWordsAndProbs = getPrevWordsAndProbs(word);
		String[] prevWords = Viterbi.concat(prevWordsAndProbs[0].getWords(), prevWordsAndProbs[1].getWords());
		double[] probs1 = prevWordsAndProbs[0].getProbs();
		double[] probs2 = prevWordsAndProbs[1].getProbs();
		double[] prevProbs = Arrays.copyOf(probs1, probs1.length + probs2.length);
		System.arraycopy(probs2, 0, prevProbs, probs1.length, probs2.length);
		double bestProb = Double.NEGATIVE_INFINITY;
		String bestPrevWord = "";
		for (int index = 0; index < prevWords.length; index++) {
		    double indexProb = getAlignedWordProb(word, i, a, sourceSentence, length) + prevProbs[index];
		    if (prevRow.containsKey(prevWords[index])) {
			indexProb += prevRow.get(prevWords[index]).getProb();
		    } else {
			indexProb += Double.NEGATIVE_INFINITY;
		    }
		    if (indexProb > bestProb) {
			bestProb = indexProb;
			bestPrevWord = prevWords[index];
		    }
		}
		TableEntry bestPrevEntry = prevRow.get(bestPrevWord);
		currentRow.put(word, new TableEntry(word, bestProb, bestPrevEntry));
	    }
	    dpTable.add(currentRow);
	}
	TableEntry bestEntry = new TableEntry("", Double.NEGATIVE_INFINITY, null);
	Hashtable<String, TableEntry> lastRow = dpTable.get(length-1);
	Enumeration en = lastRow.keys();
	while (en.hasMoreElements()) {
	    TableEntry curEntry = (TableEntry) lastRow.get(en.nextElement());
	    if (curEntry.getProb() > bestEntry.getProb()) { bestEntry = curEntry; }	    
	}
	return getBestSentence(bestEntry, length);
    }

    /**
     * Finds likely destination word translations for a word in the source sentence

     @param sourceSentence The sentence to decode
     @param a The alignment to use in decoding
     @param position The position of the word in the source sentence
     @return A string array of all likely detination words
    */

    private String[] getAllCandidates(String[] sourceSentence, Alignment a, int position) {

	String[] candidates = new String[0];

	ArrayList<String> sourceWords = new ArrayList<String>();
	int [] alignmentArray = a.getAlignment();
	for (int i = 0; i < sourceSentence.length; i++) {
	    if (alignmentArray[i] == position) {
		sourceWords.add(sourceSentence[i]);
	    }
	}
	
	for (String sourceWord : sourceWords) {
	    candidates = Viterbi.concat(candidates, translationCandidates.get(sourceWord));
	}

	return candidates;
    }

    /**
     * Reconstructs a the best destination sentence given a table entry.
     * Repeatedly follows pointers to previous table entries

     @param bestEntry A table entry. Intended to be the entry in the final column of the DP table with the highest probability
     @param length The length of the sentence to be constructed
     @return A string array representation of the reconstructed sentence
    */

    private String[] getBestSentence(TableEntry bestEntry, int length) {
	int index = length-1;
	String[] solutionArray = new String[length];
	TableEntry curEntry = bestEntry;
	while (curEntry != null) {
	    solutionArray[index] = curEntry.getE();
	    curEntry = curEntry.getPrev();
	    index -= 1;
	}
	return solutionArray;
    }

    /**
     * Calculates the translational probability of adding a word to the target sentence
     * at a specified position. This uses the alignment to determine which words in the
     * source sentence are generated by the new target word.

     @param e The target language word to add
     @param i The position in the target sentence to add the word
     @param a The alignment to use
     @param sourceSentence The source sentence
     @param destLength The overall intended length of the target sentence
     @return The probability of this addition
    */
    public double getAlignedWordProb(String e, int i, Alignment a, String[] sourceSentence, int destLength) {

	double product = 0.0; // max log prob

	int[] alignmentArray = a.getAlignment();
	
	// ensure the source sentence and the alignment array are of the same length
	assert sourceSentence.length == alignmentArray.length;

	// find the indices of the source words generated by the target word
	ArrayList<Integer> alignedIndices = new ArrayList<Integer>();
	for (int index = 0; index < alignmentArray.length; index++) {
	    if (alignmentArray[index] == i) {
		alignedIndices.add(index);
	    }
	}
	
	// multiply in translation and distortion probabilities
	for (int index : alignedIndices) {
	    String word;
	    product += params.getLogT(sourceSentence[index], e);
	    product += params.getLogD(index + 1, i, destLength, sourceSentence.length);
	}
	
	// multiply in the fertility probability
	product += params.getLogN(alignedIndices.size(), e) + Math.log10(MathUtils.factorial(alignedIndices.size()));

	return product;
    }

    /**
     * Generates a collection of words that are likely to come before a
     * specified word, and their associated bigram probabilities.

     @param e The target sentence word
     @return An array of structures containing likely words and their bigram probabilities
    */

    protected WordsAndProbabilities [] getPrevWordsAndProbs(String e) {

	WordsAndProbabilities [] wapArray = new WordsAndProbabilities[2];

	// if e is unknown
	if (unknownSourceWords.contains(e)) {
	    wapArray[0] = unknownToKnownWAP;
	    double[] unknownBigramProbs = new double[unknownSourceWords.size()];
	    int index = 0;
	    for (String unknown : unknownSourceWords) {
		// check if the pair of unknown words appeared in the source sentence
		if (unknownSourceWordPairs.contains(unknown + " " + e))
		    unknownBigramProbs[index] = 0.0; // full log probability
		else
		    unknownBigramProbs[index] = unknownBigramProb;
		index++;
	    }
	    wapArray[1] = new WordsAndProbabilities(unknownSourceWordsArray, unknownBigramProbs);
	
	// e is unknown
	} else {
	    wapArray[0] = nextToPrev.getInnerData(e);
	    wapArray[1] = knownToUnknownWAP;
	}
	return wapArray;
    }

    /**
     * Generates a collection of words that are likely word-level translations
     * of a specified word.

     @param f The source sentence word
     @param count The number of words to generate
     @return An array of the best single-word translations
    */
    public String [] getBestWords(String f, int count) {
    	PriorityQueue<Pair<String, Double>> wordsQueue = new PriorityQueue<Pair<String, Double>>(uniqueDestWords.length, new SentenceComparator());

    	int nonZeroCount = 0;

    	// stick words in with their associated probabilities
    	// to sort them
    	for(String e : uniqueDestWords) {
    		//motivation for the below:
    		//t(e|f) = t(f|e) * p(e) / p(f)
    		double wordProb = params.getLogT(f, e) + languageModel.queryNgram(e);
    		if(wordProb != Double.NEGATIVE_INFINITY) {
    			wordsQueue.add(new Pair<String, Double>(e, wordProb));
    			nonZeroCount++;
    		}
    	}

    	int numToReturn = count;
    	if(numToReturn > nonZeroCount)
    		numToReturn = nonZeroCount;

    	// get them out and put them in the array
    	String [] bestTranslations = new String[numToReturn];

    	for(int i = 0; i < bestTranslations.length; i++) {
    		Pair<String, Double> pair = wordsQueue.poll();
    		assert(pair != null);
    		bestTranslations[i] = pair.getLeft();
    	}

    	return bestTranslations;
    }
    
    /**
     * Generates all likely single-word translations for words from the source sentence. 
     * Saves them as a member variable.
     
     @param sourceSentence The source language sentence
    */    
    protected void getAllBestWords(String[] sourceSentence) {
	translationCandidates = new Hashtable<String, String[]>();
	for(String sourceWord : sourceSentence) {
		String [] bestWords = getBestWords(sourceWord, 100); // customize this
		translationCandidates.put(sourceWord, bestWords);
	}
    }

    /**
     * Sorts the list of destination word according to their unigram probabilities.
     */
    protected void setSortedDestWords() {
    	PriorityQueue<Pair<String, Double>> wordsQueue = new PriorityQueue<Pair<String, Double>>(uniqueDestWords.length, new SentenceComparator());

    	// stick things in to sort them
    	for(String e : uniqueDestWords) {
	    wordsQueue.add(new Pair<String, Double>(e, languageModel.queryNgram(e)));
    	}

    	// get them out and put them in the array
    	for(int i = 0; i < uniqueDestWords.length; i++) {
    		Pair<String, Double> pair = wordsQueue.poll();
    		assert(pair != null);
    		uniqueDestWords[i] = pair.getLeft();
    	}
    }

    public static void main(String [] args) {

	// testing code -- can be used to decode from spa > eng given a command line filename

	System.err.println("Setting up ASDecoder for Spanish >> English");
	String sourceTextFile = "../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.es";
        String destTextFile = "../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.en";
        String paramsTextFile = "../parameters/textfiles/model3-iter-patched-nosmoothing-4";
        String sbpFile = "../language_models/europarl_2-to-10_stripped.training.es-en.en.ntp.sbp";
        String srilmFile = "../language_models/europarl_2-to-10_stripped.training.es-en.en.lm";

	Model3Params params = new Model3Params(paramsTextFile);
        CorpusFromTextPair corpus = new CorpusFromTextPair(sourceTextFile, destTextFile);
        BigramsFromSrilmFile bfsf = new BigramsFromSrilmFile(srilmFile);
        SortedProbabilityAccessor nextToPrevs = new SortedProbabilityAccessor(sbpFile);
	ASDecoder decoder = new ASDecoder(params, corpus, bfsf, nextToPrevs);

	System.err.println("Setup complete.");
 
	String filename = args[0];
	System.err.println("Decoding file: "+ filename);
	
	try {
	    BufferedReader in = new BufferedReader(new FileReader(filename));
	    String sourceSentence;
	    while ((sourceSentence = in.readLine()) != null) {
		System.out.println(decoder.decode(sourceSentence));
	    }
	    in.close();
	} catch (IOException e) {
	    System.err.println("Failed to read from: " + filename);
	}
    }
}

/**
 * A class to hold the necessary data for an entry in the DP table
 */
class TableEntry {
    
    private String e;
    private double prob;
    private TableEntry prev;

    public TableEntry(String word, double probability, TableEntry previous) {
	e = word;
	prob = probability;
	prev = previous;
    }

    public String getE() { return e; }
    public double getProb() { return prob; }
    public TableEntry getPrev() { return prev; }
}
