import java.util.*;
import java.io.*;
import java.text.*;


/**
 * Implements the core functionality of a dynamic programming algorithm for decoding
 * IBM models 1-3. Certain methods are intended to be overridden in subclasses in order
 * to achieve model-specific functionality.
 */

class DPDecoder implements Decoder {

    // sources of probabilities and corpus information
    protected SortedProbabilityAccessor nextToPrevs;
    protected ParameterAccessor params;
    protected CorpusAccessor corpus;
    protected LanguageModel languageModel;


    // configuration information
    protected boolean handleUnknownWords;
    protected boolean requireFullCoverage;

    protected boolean useStartingBigrams;  //whether to use "<s> word" bigrams
    protected boolean useEndingBigrams;  //whether to use "word </s>" bigrams

    protected double unigramWeight;  //amount to mix in unigrams with LM bigram values

    //max # of highest-unigram dest words in outerLoopWords if restrictOuterLoopWords
    protected int maxCommonOuterWords;
    //max # words to be added per source sentence word if restrictOuterLoopWords
    protected int outerWordsPerSourceWord;
    
    protected boolean restrictOuterLoopWords;
    protected boolean restrictCoverage; //limit possible c vals
    protected boolean restrictLastCovered; //limit possible j vals
    protected boolean restrictFertility; //limit possible l vals (computeNonSkipQElement)
    protected boolean ignoreLastCovered; //ignore the j part of the DP table indexing; supercedes restrictLastCovered


    // information culled from probabilities/corpus objects
    protected String[] uniqueDestWords;
    protected String[] sortedDestWords;

    protected String preprocessedSourceSentence;

    // cached data for unknown-words handling
    protected boolean hasUnknownWords;
    protected String[] unknownSourceWords; 
    protected HashSet<String> unknownSourceWordSet;
    protected HashSet<String> unknownSourceWordPairs;
    protected WordsAndProbabilities knownToUnknownWAP;
    protected WordsAndProbabilities unknownToKnownWAP;

    protected HashSet<String> outerLoopWordSet;
    protected String[] outerLoopWords;

    
    // default probabilities for unknown words

    // log probability of an unknown word following a known word and vice versa
    protected static double unknownBigramProb = -6;
    protected static double unknownUnigramProb = -5;

    
    // output options
    protected boolean showAlignment;  //whether results include alignment info
    protected boolean verbose;


    /**
     * Constructs and initializes the object given the provided corpus and translation
     * parameters.

     @param p the parameter accessor
     @param c the corpus accessor
     @param ntp the sorted probability accessor for optimization
     */

    public DPDecoder(ParameterAccessor p, CorpusAccessor c, LanguageModel lm, SortedProbabilityAccessor ntp) {

    	// the contents of this function (should) align with the
    	// list of member variables above

    	// sources of probabilities and corpus information
        nextToPrevs = ntp;
        params = p;
        corpus = c;
        languageModel = lm;
        
        // configuration information
        handleUnknownWords = true;
        requireFullCoverage = true; //this is overridden by DPModel3

        useStartingBigrams = true;
        useEndingBigrams = true;

        unigramWeight = 1.0;

        maxCommonOuterWords = 200;
        outerWordsPerSourceWord = 20;

        restrictOuterLoopWords = true;
        restrictCoverage = true;
        restrictLastCovered = true;
        restrictFertility = true;
        ignoreLastCovered = true;
        
        // information culled from probabilities/corpus objects
        uniqueDestWords = corpus.getWordList(true);
        //sortedDestWords must be set after unknown* stuff

        preprocessedSourceSentence = "";
        
        // cached data for unknown-words handling
        String [] tempStrings = new String[0];
        double [] tempDoubles = new double[0];

        hasUnknownWords = false;
        unknownSourceWords = new String[0];
        unknownSourceWordSet = new HashSet<String>();
        unknownSourceWordPairs = new HashSet<String>();
        knownToUnknownWAP = new WordsAndProbabilities(tempStrings, tempDoubles);
        unknownToKnownWAP = new WordsAndProbabilities(tempStrings, tempDoubles);

        outerLoopWordSet = new HashSet<String>();
        outerLoopWords = uniqueDestWords;

        setSortedDestWords();

        // output options

	showAlignment = false; //include alignment information in decoded sentences
	verbose = false; //for now, always print verbose
    }

    /**
     * Sets the sortedDestWords member variable based upon the
     * current parameters object being used by the class.
     * This member variable is the array uniqueDestWords
     * sorted in descending order of the unigram probabilities
     * of the words in the array.
     */

    protected void setSortedDestWords() {
    	PriorityQueue<Pair<String, Double>> wordsQueue = new PriorityQueue<Pair<String, Double>>(uniqueDestWords.length, new SentenceComparator());

    	// stick things in to sort them
    	for(String e : uniqueDestWords) {
    		wordsQueue.add(new Pair<String, Double>(e, getUnigramProb(e)));
    	}

    	// get them out and put them in the array
    	sortedDestWords = new String[uniqueDestWords.length];

    	for(int i = 0; i < sortedDestWords.length; i++) {
    		Pair<String, Double> pair = wordsQueue.poll();
    		assert(pair != null);
    		sortedDestWords[i] = pair.getLeft();
    	}
    }

    /**
     * Returns an array containing up to a specified number of
     * the destination-language words most likely to be
     * translated to from the given source-language word.

     @param f the word to be translated
     @param count the number of desired results
     @return up to the desired number of the
     * destination-language words most likely to be translated
     * to from the given source-language word
     */

    protected String [] getBestTranslations(String f, int count) {
    	PriorityQueue<Pair<String, Double>> wordsQueue = new PriorityQueue<Pair<String, Double>>(uniqueDestWords.length, new SentenceComparator());

    	int nonZeroCount = 0;

    	// stick words in with their associated probabilities
    	// to sort them
    	for(String e : uniqueDestWords) {
    		//motivation for the below:
    		//t(e|f) = t(f|e) * p(e) / p(f)
    		double wordProb = getTranslationProb(f, e) + getUnigramProb(e);
    		if(wordProb != Double.NEGATIVE_INFINITY) {
    			wordsQueue.add(new Pair<String, Double>(e, wordProb));
    			nonZeroCount++;
    		}
    	}

    	int numToReturn = count;
    	if(numToReturn > nonZeroCount)
    		numToReturn = nonZeroCount;

    	// get them out and put them in the array
    	String [] bestTranslations = new String[numToReturn];

    	for(int i = 0; i < bestTranslations.length; i++) {
    		Pair<String, Double> pair = wordsQueue.poll();
    		assert(pair != null);
    		bestTranslations[i] = pair.getLeft();
    	}

    	return bestTranslations;
    }

    /**
     * Adds the provided words to the set of outer loop words.
     * These changes must be manually ported over to the
     * outerLoopWords array.
     */

    protected void addToOuterSet(String [] otherWords) {
    	for(String s : otherWords) {
    		outerLoopWordSet.add(s);
    	}
    }

    // --- functions that can be overridden to change
    //     commonly-customized aspects of the decoding process
    //     (e.g. how to perform the "inner" and "outer" loops over
    //     destination words, how to obtain values for alignment and
    //     fertility probabilities, etc.)
    
    /* Optimization and result quality helper functions */


    /**
     * Configures the object to handle words in the given sentence that do not appear
     * in the corpus. A number of member variables are updated to reflect the unknown
     * words. This method is called automatically opon decoding.

     @param sourceSentence sentence to check for unknown words
     @return whether any unknown words were found
     */

    protected boolean detectUnknownWords(String sourceSentence) {
	String[] uniqueSourceWords = corpus.getWordList(false);
	String[] splitSourceSentence = sourceSentence.split("\\s+");
	unknownSourceWordSet = new HashSet<String>();

	// find unknown words
	for (String word : splitSourceSentence) {
	    boolean isKnown = false;
	    for (String knownWord : uniqueSourceWords) {
		if (word.equals(knownWord)) {
		    isKnown = true;
		    break;
		}
	    }
	    if (! isKnown)
		unknownSourceWordSet.add(word);
	}
	
	// make array of unknown words
	unknownSourceWords = new String[unknownSourceWordSet.size()];
	int index = 0;
	for (String uw : unknownSourceWordSet) {
	    unknownSourceWords[index] = uw;
	    index++;
	}
	
	// setup WAPs for bigram lookup
	double[] lowProbs = new double[unknownSourceWords.length];
	for (int i = 0; i < lowProbs.length; i++) {lowProbs[i] = unknownBigramProb;}
	knownToUnknownWAP = new WordsAndProbabilities(unknownSourceWords, lowProbs);
	
	lowProbs = new double[uniqueDestWords.length];
	for (int i = 0; i < lowProbs.length; i++) {lowProbs[i] = unknownBigramProb;}
	unknownToKnownWAP = new WordsAndProbabilities(uniqueDestWords, lowProbs);

	printVerbose("Found unknown words: " + Arrays.toString(unknownSourceWords));
	
	// construct outer loop words as uniqueDestWords + unknown words
	addToOuterSet(unknownSourceWords);

	// find unknown source word pairs
	String prevWord = "";
	unknownSourceWordPairs = new HashSet<String>();
	for (String word : splitSourceSentence) {
	    boolean isUnknown = false;
	    for (String unknownWord : unknownSourceWords) {
		if (word.equals(unknownWord))
		    isUnknown = true;
	    }
	    if (! prevWord.equals("") && isUnknown) {
		unknownSourceWordPairs.add(prevWord + " " + word);
	    }
	    if (isUnknown)
		prevWord = word;
	    else
		prevWord = "";
	}
	printVerbose("Unknown word pairs:");
	for (String u : unknownSourceWordPairs) {
	    printVerbose("\t"+u);
	}

	hasUnknownWords = (unknownSourceWords.length != 0);
	return hasUnknownWords;
    }

    /**
     * Appends to outerLoopWords the words that are most likely
     * to be translated to from each of the words in the specified
     * source sentence.

     @param sourceSentence the source sentence
     */

    protected void appendSentenceRelevantWords(String sourceSentence) {
	String[] splitSourceSentence = sourceSentence.split("\\s+");

	// grab top unigram words
	int numTopWords = maxCommonOuterWords;
	if(numTopWords > sortedDestWords.length)
		numTopWords = sortedDestWords.length;
	
	String [] topWords = new String [numTopWords];
	for(int i = 0; i < numTopWords; i++) {
		topWords[i] = sortedDestWords[i];
	}

	addToOuterSet(topWords);

	for(String sourceWord : splitSourceSentence) {
		String [] bestWords = getBestTranslations(sourceWord, outerWordsPerSourceWord);
		addToOuterSet(bestWords);
	}
    }

    /**
     * Allows the decoder to configure its state based upon
     * the provided sentence. This is called from within the
     * decoder at the start of each decode to set various
     * member variables depending on how the decoder is configured.
     */

    protected void preprocessSentence(String sourceSentence) {

	if (sourceSentence.equals(preprocessedSourceSentence )) {
	    return;
	}
	preprocessedSourceSentence = sourceSentence;

    	// reset outerLoopWordSet
    	outerLoopWordSet = new HashSet<String>();
    	if(!restrictOuterLoopWords) {
    		addToOuterSet(uniqueDestWords);
    	} else {
    		appendSentenceRelevantWords(sourceSentence);
    	}

    	// sets member variables and adds to outerLoopWordSet
    	detectUnknownWords(sourceSentence);

    	// transfer outer loop words
    	outerLoopWords = new String[outerLoopWordSet.size()];

    	int index = 0;
    	for(String s : outerLoopWordSet) {
    		assert(index < outerLoopWordSet.size());
    		outerLoopWords[index] = s;
    		index++;
    	}
    }

    /**
     * Determines what words are used to fill out the next level of the dynamic programming
     * table.
     
     @param splitSourceSentence a string array representation of the input sentence
     @param level the level to be filled in
     @param destLength the final length of the decoded sentence (the final level)
     @return an array of detination language words whose positions in the table should be filled
     */

    protected String [] getNextWords(String [] splitSourceSentence, int level, int destLength) {
 	return outerLoopWords; 
    }

    /**
     * Determine what lengths should be considered for the decoded sentence.

     @param sourceSentence the sentence to be decoded
     @return a pair of integers representing the minimum and maximum sentence lengths
     */

    protected Pair<Integer, Integer> getDestLengthRange(String sourceSentence) {
        int length = sourceSentence.split("\\s+").length;
        int min = length / 2;
        if(min < 1)
        	min = 1;
	return new Pair<Integer, Integer>(min, length + 2); //TEMP for testing
    }

    /**
     * Determines the range of allowable amounts of source-sentence coverage
     * for hypothesis sentences at a given level of the dynamic-programming table.

     @param splitSourceSentence the source sentence, in array-of-words form
     @param level the level of the table being considered (the length of
     * hypothesis sentences)
     @param length the final desired length for the translation sentences
     @return a pair indicating the lowest and highest c value for which entries
     * in the dynamic-programming table should be created
     */

    protected Pair<Integer, Integer> getCRange(String [] splitSourceSentence, int level, int length) {
    	int I = length;
    	int J = splitSourceSentence.length;

    	// fill in all positions by default:
    	int cLow = 0;
    	int cHigh = J;

    	if(restrictCoverage) {
    		// trying to be smart:
    		int cStarter = (level * J) / I;

    		cLow = cStarter - J/I - 1;
    		if(!requireFullCoverage) {
    			cLow -= J/4; //allow about 1/4 of words to be spurious
    		}
    		if(cLow < 0)
    			cLow = 0;
    		cHigh = cStarter + 3;
    		if(cHigh > J)
    			cHigh = J;
    	}

    	//printVerbose("cLow=" + Integer.toString(cLow) + ", cHigh=" + Integer.toString(cHigh));

    	return new Pair<Integer, Integer>(cLow, cHigh);
    }

    /**
     * Determines the range of allowable positions in the source sentence to
     * be covered last by hypothesis sentences at a given level of the
     * dynamic-programming table.

     @param c the number of words covered in the source sentence by the
     * hypothesis sentences for which the returned range of allowable positions
     * is enforced
     @param splitSourceSentence the source sentence, in array-of-words form
     @param level the level of the table being considered (the length of
     * hypothesis sentences)
     @param length the final desired length for the translation sentences
     @return a pair indicating the lowest and highest j value for which entries
     * in the dynamic-programming table with the given c value should be created
     */

    protected Pair<Integer, Integer> getJRange(int c, String [] splitSourceSentence, int level, int length) {
    	int I = length;
    	int J = splitSourceSentence.length;

    	int jStarter = (level * J) / I;

    	// fill in all positions by default:
    	int jLow = 1;
    	int jHigh = J;

    	if(ignoreLastCovered || c == 0) {
    		// there is no "last covered index"
    		jLow = jHigh = -1;
    	} else if(restrictLastCovered) {
    		// trying to be smart:
    		jLow = jStarter - J/I;
    		if(jLow < 1)
    			jLow = 1;
    		jHigh = jStarter + 3;
    		if(jHigh > J)
    			jHigh = J;
    	}

    	//printVerbose("jLow=" + Integer.toString(jLow) + ", jHigh=" + Integer.toString(jHigh));

    	return new Pair<Integer, Integer>(jLow, jHigh);
    }

    /**
     * Returns the maximum fertility that should be considered
     * for creating the given entry in the dynamic-programming
     * table. The value returned should still be possible as
     * a fertility for the last word in the hypothesis sentence
     * at this position in the DP table; that is, it should not
     * exceed j or c.

     @param c the first index to the DP table
     @param i the second index to the DP table
     @param j the third index to the DP table
     @param e the fourth index to the DP table
     @param sourceSentence the source sentence
     @param destLength the final length for the hypothesis sentences
     @return the maximum fertility to be considered
     */

    protected int getFertilityMax(int c, int i, int j, String e, String [] sourceSentence, int destLength) {
    	// consider all fertilities by default:
    	int fertMax = sourceSentence.length;

    	if(restrictFertility) {
    		// trying to be clever
    		fertMax = 3;
    	}

    	if(fertMax > c)
    		fertMax = c;
    	if(fertMax > j)
    		fertMax = j;

    	return fertMax;
    }

    /**
     * Returns an array of objects indicating the possible preceding words
     * and associated bigram probabilities for the given destination language
     * word, given that this word is the first word in the sentence. The only
     * meaningful entries in the returned objects will be those for which the
     * indicated preceding word is the SRILM start-of-sentence indicator;
     * the full flexibility of returning an array of WordsAndProbabilities
     * objects is provided for the case where subclasses find it more
     * conventient to work exclusively with such arrays.

     @param e the word at the start of the sentence
     @return an array of WordsAndProbabilities objects, conveying the
     * information about the possible predecessor words and
     * their probabilities
     */

    protected WordsAndProbabilities [] getSentenceStartWAP(String e) {
    	String [] loopWords = new String[1];
    	loopWords[0] = "<s>";
    	double [] loopProbs = new double[1];
    	loopProbs[0] = 0.0; //maximum probability

    	if(useStartingBigrams) {
    		if(unknownSourceWordSet.contains(e)) {
    			loopProbs[0] = unknownBigramProb;
    		} else {
    			// look up unigram
    			double unigramProb = languageModel.queryNgram(e);

    			if(unigramProb < -100) {
    				// return a smoothed probability for the word;
    				// essentially, a smoothing of its unigram prob
    				loopProbs[0] = -4;
    			} else {
    				// look up bigram
    				double bigramProb = languageModel.queryNgram("<s> " + e);

    				if(bigramProb < -100) {
    					//smooth the bigram probability
    					bigramProb = -4;
    				}

    				double mixCoeff = 1.0 / (1 - unigramProb);

    				loopProbs[0] = mixCoeff * bigramProb + (1 - mixCoeff) * unigramProb;
    			}
    		}
    	}

    	WordsAndProbabilities [] wapArray = new WordsAndProbabilities[1];
    	wapArray[0] = new WordsAndProbabilities(loopWords, loopProbs);
    	return wapArray;
    }

    /**
     * Returns an array of objects containing the destination language words that
     * could precede a given word and the associated bigram probabilities.

     @param e the word whose possible predecessors will be found
     @param i the position of e in the decoded sentence
     @return an array of WordsAndProbabilities objects, each containing arrays
     * of previous words and bigram probabilities
     */

    protected WordsAndProbabilities [] getPrevWordsAndProbs(String e, int i) {
        if(i == 1) {
            return getSentenceStartWAP(e);
        } else {
	    //return array of two WAP objects
	    
	    WordsAndProbabilities [] wapArray = new WordsAndProbabilities[2];

	    // e is unknown
	    if (unknownSourceWordSet.contains(e)) {
		wapArray[0] = unknownToKnownWAP;
		double[] unknownBigramProbs = new double[unknownSourceWords.length];
		for (int j = 0; j < unknownBigramProbs.length; j++) {
		    // check if the pair of unknown words appeared in the source sentence
		    if (unknownSourceWordPairs.contains(unknownSourceWords[j] + " " + e))
			unknownBigramProbs[j] = 0.0; // full log probability
		    else
			unknownBigramProbs[j] = unknownBigramProb;
		}
		wapArray[1] = new WordsAndProbabilities(unknownSourceWords, unknownBigramProbs);
	    } else {
		//wapArray[0] = getKnownPrevWordsAndProbs(e, i); //CUSTOMIZE_UNKNOWN
		wapArray[0] = nextToPrevs.getInnerData(e);
		wapArray[1] = knownToUnknownWAP;
	    }
	    return wapArray;
        }
    }

    /**
     * Returns the probability that the given word is seen at the end of
     * a sentence, along with any other modifications to the final
     * probability of a hypothesis sentence that can only be
     * determined once the sentence is complete. This function
     * may be overridden to apply further modifications to the
     * probability of a completed sentence; see the
     * DPModel3Decoder subclass for an example.

     @param e the last word of a sentence
     @param sourceSentence the source sentence, as an array of words
     @param coverageBitField an integer expressing which positions
     * of the source sentence have been covered by the translation
     * ending in the provided word
     @return the probability of the bigram consisting of the given word
     * and the end-of-sentence indicator
     */

     protected double getSentenceEndProb(String e, String [] sourceSentence, int coverageBitField) {
     	if(requireFullCoverage && (coverageBitField + 1) != (1 << sourceSentence.length)) {
     		// we require full coverage, but this didn't
     		// reach full coverage
     		return Double.NEGATIVE_INFINITY;
     	}
     	if(!useEndingBigrams) {
     		return 0.0; //full log probability
     	}

	if(unknownSourceWordSet.contains(e)) {
		return unknownBigramProb;
	}

    	double endProb = 0.0;

    	// look up unigram
    	double unigramProb = languageModel.queryNgram(e);

    	if(unigramProb < -100) {
    		// return a smoothed probability for the word;
    		// essentially, the prob that this word shows up
    		// at the end of a sentence
    		endProb = -3;
    	} else {
    		// look up bigram
    		double bigramProb = languageModel.queryNgram(e + " </s>");
    		double defaultBigramProb = -2.5;

    		if(bigramProb < -100) {
    			//smooth the bigram probability
    			endProb = -2.5;
    		}
    		else {
    			//double mixCoeff = 1.0 / (1 - unigramProb);
    			double reverseMixCoeff = 1.0 / (unigramProb - 1);
    			double mixCoeff = 1 - reverseMixCoeff * reverseMixCoeff;

    			endProb = mixCoeff * bigramProb + (1 - mixCoeff) * defaultBigramProb;
    		}
    	}

    	return endProb;
     }

    /**
     * Returns the probability of the given unigram according to the language model.

     @param unigram the unigram to be looked up
     @return the probability of the unigram
     */

    protected double getUnigramProb(String unigram) {
    	if(unknownSourceWordSet.contains(unigram)) {
    		return unknownUnigramProb;
    	}
    	return languageModel.queryNgram(unigram);
    }

    /**
     * Mixes the given bigram and unigram probabilities. Depending on the implementation
     * here or in subclasses, this mixing may be for the purposes of applying smoothing,
     * punishing words with low unigram values, or something else.

     @param bigramProb the bigram probability
     @param unigramProb the unigram probability
     @return the mixed probability resulting from the two probabilities
     */

    protected double mixLMProbs(double bigramProb, double unigramProb) {
    	//double bigramWeight = 1.0 - unigramWeight;//0.5;

    	//double weightedAverage = logTen(bigramWeight * Math.pow(10, bigramProb) + (1 - bigramWeight) * Math.pow(10, unigramProb));
    	double multiplied = bigramProb + unigramWeight * unigramProb;

    	return multiplied;
    }

    /**
     * Returns the given probability adjusted to be comparable to decoded sentences of
     * other lengths.

     @param sourceSentence the input sentence
     @param destSentence the decoded sentence
     @param prob the raw probability
     @return the adjusted probability
     */
    
    protected double getAugmentedProbability(String sourceSentence, String destSentence, double prob) {
    	int sentenceLength = destSentence.split("\\s+").length;
    	int sourceLength = sourceSentence.split("\\s+").length;

    	int lengthDiff = Math.abs(sourceLength - sentenceLength);

    	// original strategy -- overly favors long sentences
    	//return prob / sentenceLength;

    	// favor things close to the length of the other-language sentence
    	//return prob - 2.0 * lengthDiff * lengthDiff;

    	// attempting to penalize too-short sentences more heavily, since they'll
    	// have a lower probability than too-long sentences
    	if(sentenceLength <= sourceLength) {
    		return prob - 2.0 * lengthDiff * lengthDiff;
    	} else {
    		return prob - 0.5 * lengthDiff;
    	}
    }
    
    /* End optimization/quality-related helper functions */

    /* Model-specific helper functions. To be overridden by subclasses. */

    /**
     * Returns the probability that one known word is generated by another
     * known word under the generative model. This may be overridden to
     * allow for secondary features such as an alternate source of
     * probability values.

     @param f the known translation-source language word
     @param e the known translation-destination language word
     @return the probability for the given pair of words
     */

    protected double getKnownTranslationProb(String f, String e) {
    	return params.getLogT(f, e);
    }

    /**
     * Returns the probability that one word is generated by another word
     * under the generative model.

     @param f the translation-source language word
     @param e the raw translation-destination language word
     @return the probability for the given pair of words
     */

    protected double getTranslationProb(String f, String e) {
    	// if both f and e are unknowns and match, give full prob
    	// otherwise if one or the other is unknown, give -infty

	if (unknownSourceWordSet.contains(f) || unknownSourceWordSet.contains(e)) {
	    if (e.equals(f))
		return 0.0; // full log prob
	    else
		return Double.NEGATIVE_INFINITY;
	} else {
	    return getKnownTranslationProb(f, e);
	}
    }

    /**
     * Returns the probability that a known translation destination language word
     * generates a specific number of source language words. This method is called
     * by the main getFertilityProb function in the case that the input word
     * is known.

     @param l the number of generated words
     @param e the known destination language word
     @return the fertility probability
     */

    protected double getKnownFertilityProb(int l, String e) {
    	return 0.0;
    }

    /**
     * Returns the probability that a translation destination language word
     * generates a specific number of source language words.

     @param l the number of generated words
     @param e the destination language word
     @return the fertility probability
     */

    protected double getFertilityProb(int l, String e) {
    	// unknown will get 0.0 for l=1, -infinity otherwise

	if (unknownSourceWordSet.contains(e)) {
	    if (l == 1)
		return 0.0; // full log prob
	    else
		return Double.NEGATIVE_INFINITY;
	} else {
	    return getKnownFertilityProb(l, e);
	    //return 0.0;
	}
    }

    /**
     * Returns the probability of an alignment occurring between two
     * specific positions in sentences of given lengths.

     @param i the position in the translation-destination sentence
     @param j the position in the translation-source sentence
     @param I the length of the translation-destination sentence
     @param J the length of the translation-source sentence
     @return the alignment probability
     */

    protected double getAlignmentProb(int i, int j, int I, int J) {
    	return params.getLogA(i, j, I, J); 
    }

    /* End Model-specific helper functions */

    /**
     * Produces a bitfield representing the new coverage in a sentence
     * caused by a word with a given alignment and fertility.
     * The nature of this encoding restricts the decoder to handling
     * sentences of length at most equal to the number of bits in
     * an integer (32).

     @param j the position in the translation-source sentence of the
     * last word covered
     @param l the fertility of the coverage
     @return an integer representing this coverage
     */

    protected int coverageToBitField(int j, int l)
    {
        int output = 0;
        for(int k = j - l + 1; k <= j; k++)
        {
            output += (1 << (k - 1));
        }
        return output;
    }

    /**
     * Returns the highest-probability translation of a given length for
     * the provided sentence, along with its probability.

     @param sourceSentence the sentence to be translated
     @param length the length of the resulting translation
     @return a pair of the translation and its probability, or a pair of
     * the empty string and negative infinity if the sentence cannot be
     * translated to this length
     */
    
    public Pair<String, Double> decodeToLength(String sourceSentence, int length) {
	ArrayList<Pair<String, Double>> result = decodeToLength(sourceSentence,
								length, 1);
	if (result.size() == 0)
	    return new Pair<String, Double>("", Double.NEGATIVE_INFINITY);
	
	return result.get(0);
    }

    /**
     * Returns the highest-probability translations of a given length for
     * the provided sentence, along with their probabilities.

     @param sourceSentence the sentence to be translated
     @param length the length of the resulting translation
     @param maxResults the maximum number of resulting translations to
     * return
     @return an ArrayList of translation/probability pairs of length
     * up to maxResults, depending on the number of translations found
     */

    public ArrayList<Pair<String, Double>> decodeToLength(String sourceSentence, int length, int maxResults)
    {
	preprocessSentence(sourceSentence);
	if(hasUnknownWords && !handleUnknownWords) {
		// we aren't handling unknown words, so report finding no results
		return new ArrayList<Pair<String, Double>>();
	}
	
        //configure base case
        QLevel previousLevel = new QLevel(0);
        previousLevel.addElement(new QElement(0.0, 0, -1, "<s>", 0, null)); //full probability
        
        String [] splitSourceSentence = sourceSentence.split("\\s+");
        int I = length;                      //length of the translation-dest sentence
        int J = splitSourceSentence.length;  //length of the translation-source sentence

        //build up length of the hypothesis dest-language sentences
        for (int level = 1; level <= length; level++) {
            printVerbose("Starting level " + Integer.toString(level) + " of " + Integer.toString(length) + "...");

            QLevel currentLevel = new QLevel(level);

            Pair<Integer, Integer> cRange = getCRange(splitSourceSentence, level, length);
            int cLow = cRange.getLeft();
            int cHigh = cRange.getRight();
            for(int cPrime = cLow; cPrime <= cHigh; cPrime++) {
            	Pair<Integer, Integer> jRange = getJRange(cPrime, splitSourceSentence, level, length);
            	int jLow = jRange.getLeft();
            	int jHigh = jRange.getRight();
            	for(int jPrime = jLow; jPrime <= jHigh; jPrime++) {
            		String [] nextWords = getNextWords(splitSourceSentence, level, length);
            		for(String e : nextWords) {
            			currentLevel.addElement(computeQElement(cPrime, level, jPrime, e, previousLevel, splitSourceSentence, length));
            		}
            	}
            }

            previousLevel = currentLevel;
        }

        //print out the top N sentences

        QElement [] maxElements = previousLevel.getMaxElements(maxResults);
        ArrayList<Pair<String, Double>> results = new ArrayList<Pair<String, Double>>();

        for(int i = 0; i < maxElements.length; i++) {
        	QElement maxElement = maxElements[i];
        	if (maxElement.getProb() == Double.NEGATIVE_INFINITY) {
        		results.add(new Pair<String, Double>("", Double.NEGATIVE_INFINITY));
        	} else {
        		results.add(new Pair<String, Double>(maxElement.reconstructSentence(showAlignment), new Double(maxElement.getProb())));
        	}
        }

        NumberFormat shortDouble = NumberFormat.getInstance();
        shortDouble.setMaximumFractionDigits(4);

        int numToPrint = results.size();
        if(numToPrint > 10) {
        	numToPrint = 10;
        }

        printVerbose("Showing best " + Integer.toString(numToPrint) + " of "
        	+ Integer.toString(results.size())
        	+ " viable sentences found...");
        printVerbose("(" + Integer.toString(previousLevel.size()) + " elements at this level)");
        for(int i = 0; i < numToPrint; i++) {
        	Pair<String, Double> printPair = results.get(i);
        	printVerbose(shortDouble.format(printPair.getRight()) + ": "
        		+ printPair.getLeft());
        }

        return results;
    }

    /**
     * Returns the best translation of the provided sentence, or the
     * empty string if no translation can be found.

     @param sourceSentence the sentence to be translated
     @return the resulting translation, or the empty string if no
     * translation can be found
     */
    
    public String decode(String sourceSentence) {
        String [] resultArray = decode(sourceSentence, 1);
        if(resultArray.length == 0) {
        	return "";
        }
        return resultArray[0];
    }

    /**
     * Returns the best translations of the provided sentence.

     @param sourceSentence the sentence to be translated
     @param maxResults the maximum number of translations to return
     @return an array of the best translations of length up
     * to maxResults, depending on how many translations are found
     */

    public String [] decode(String sourceSentence, int maxResults) {
	ArrayList<Pair<String, Double>> result = decodeWithProbs(sourceSentence, maxResults);
	String[] outputStrings = new String[result.size()];
	for (int i = 0; i < outputStrings.length; i++) {
	    outputStrings[i] = result.get(i).getLeft();
	}
	return outputStrings;
    }

    /**
     * Returns the best translations of the provided sentence along
     * with their probabilities.

     @param sourceSentence the sentence to be translated
     @param maxResults the maximum number of translations to return
     @return an ArrayList of the best translations of length up
     * to maxResults (depending on how many translations are found)
     * along with their probabilities
     */

    public ArrayList<Pair<String, Double>> decodeWithProbs(String sourceSentence, int maxResults) {
        Pair<Integer, Integer> minAndMax = getDestLengthRange(sourceSentence);
        printVerbose("Decoding to lengths " + minAndMax.getLeft() + " through " + minAndMax.getRight());
        int min = minAndMax.getLeft();
        int max = minAndMax.getRight();
        PriorityQueue<Pair<String, Double>> resultsQueue = new PriorityQueue<Pair<String, Double>>(maxResults, new SentenceComparator());
        for (int i = min; i <= max; i++) {
        	ArrayList<Pair<String, Double>> newResults = decodeToLength(sourceSentence, i, maxResults);
        	for(Pair<String, Double> result : newResults) {
        		resultsQueue.add(new Pair<String, Double>(result.getLeft(), getAugmentedProbability(sourceSentence, result.getLeft(), result.getRight())));
        	}
        }

        NumberFormat shortDouble = NumberFormat.getInstance();
        shortDouble.setMaximumFractionDigits(4);

        int outputSize = resultsQueue.size();
        if(outputSize > maxResults)
        	outputSize = maxResults;
        ArrayList<Pair<String, Double>> output = new ArrayList<Pair<String, Double>>();
	printVerbose("Best " + Integer.toString(outputSize) + " overall results:");
        for(int i = 0; i < outputSize; i++) {
	    Pair<String, Double> result = resultsQueue.poll();
	    printVerbose("\t" + shortDouble.format(result.getRight()) + ": " + result.getLeft());
	    output.add(result);
        }
        return output;
    }

    /**
     * Produces the appropriate element to be placed in the specified position
     * of the dynamic-programming table.

     @param c the number of covered words; part of the indexing of the
     * dynamic-programming table
     @param i the position of the last word in the candidate translation;
     * part of the indexing of the dynamic-programming table
     @param j the position of the last word covered in the source sentence;
     * part of the indexing of the dynamic-programming table
     @param e the last word in the candidate translation;
     * part of the indexing of the dynamic-programming table
     @param previousLevel an object containing all the entries in the previous
     * level of the dynamic-programming table
     @param sourceSentence the sentence to be translated
     @param destLength the final length to be reached by the candidate
     * translations
     @return the entry for the given position in the dynamic-programming table
     */
    
    protected QElement computeQElement(int c, int i, int j, String e, QLevel previousLevel, String [] sourceSentence, int destLength) {
        QElement skipElement = computeSkipQElement(c, i, j, e, previousLevel, sourceSentence, destLength);
        if(c == 0)
            return skipElement;
        QElement nonSkipElement;
        if(!ignoreLastCovered) {
        	nonSkipElement = computeNonSkipQElement(c, i, j, e, previousLevel, sourceSentence, destLength);
        } else {
        	//call the non-skip function that maximizes over j internally
        	nonSkipElement = computeNoJNonSkipQElement(c, i, e, previousLevel, sourceSentence, destLength);
        }
        if (skipElement.getProb() > nonSkipElement.getProb())
        	return skipElement;
        else
        	return nonSkipElement;
    }

    /**
     * Produces the appropriate element to be placed in the specified position
     * of the dynamic-programming table, under the assumption that the next
     * word added to the candidate translation does not cover any words in
     * the source sentence.

     @param c the number of covered words; part of the indexing of the
     * dynamic-programming table
     @param i the position of the last word in the candidate translation;
     * part of the indexing of the dynamic-programming table
     @param j the position of the last word covered in the source sentence;
     * part of the indexing of the dynamic-programming table
     @param e the last word in the candidate translation;
     * part of the indexing of the dynamic-programming table
     @param previousLevel an object containing all the entries in the previous
     * level of the dynamic-programming table
     @param sourceSentence the sentence to be translated
     @param destLength the final length to be reached by the candidate
     * translations
     @return the entry for the given position in the dynamic-programming table,
     * under the assumption that no new source sentence words are covered
     * at this level
     */
    
    protected QElement computeSkipQElement(int c, int i, int j, String e, QLevel previousLevel, String [] sourceSentence, int destLength) {
    	double unigramProb = getUnigramProb(e);
    	double fertilityProb = getFertilityProb(0, e);
        QElement bestElement = new QElement(Double.NEGATIVE_INFINITY, c, j, e, 0, null);
        WordsAndProbabilities [] wapArray = getPrevWordsAndProbs(e, i);
        for(WordsAndProbabilities wap : wapArray) {
	        String [] loopWords = wap.getWords();
	        double [] loopProbs = wap.getProbs();
	        for (int index = 0; index < loopWords.length; index++) {
	            String prevWord = loopWords[index];
	            QElement prev = previousLevel.getElement(c, j, prevWord);
	            double newWordProb = mixLMProbs(loopProbs[index], unigramProb) + fertilityProb;
	            double newProb = newWordProb + prev.getProb();
	            if(i == destLength) {
	            	// include the end-of-sentence bigram
	            	newProb += getSentenceEndProb(e, sourceSentence, prev.getCoverage());
	            }
	            if (newProb > bestElement.getProb()) {
	                bestElement = new QElement(newProb, c, j, e, prev.getCoverage(), prev);
	            }
	        }
        }
        return bestElement;
    }

    /**
     * Produces the appropriate element to be placed in the specified position
     * of the dynamic-programming table, under the assumption that the
     * next word added to the candidate translation must cover at least
     * one word in the source sentence.

     @param c the number of covered words; part of the indexing of the
     * dynamic-programming table
     @param i the position of the last word in the candidate translation;
     * part of the indexing of the dynamic-programming table
     @param j the position of the last word covered in the source sentence;
     * part of the indexing of the dynamic-programming table
     @param e the last word in the candidate translation;
     * part of the indexing of the dynamic-programming table
     @param previousLevel an object containing all the entries in the previous
     * level of the dynamic-programming table
     @param sourceSentence the sentence to be translated
     @param destLength the final length to be reached by the candidate
     * translations
     @return the entry for the given position in the dynamic-programming table,
     * under the assumption that the last word of the candidate translation
     * must cover at least one word in the source sentence
     */
    
    protected QElement computeNonSkipQElement(int c, int i, int j, String e, QLevel previousLevel, String [] sourceSentence, int destLength) {
        double unigramProb = getUnigramProb(e);
        QElement bestL = new QElement(Double.NEGATIVE_INFINITY, c, j, e, 0, null);
        
        //for loop that produces best l
        int maxL = getFertilityMax(c, i, j, e, sourceSentence, destLength);
        for(int l = 1; l <= maxL; l++) {
            double product = getFertilityProb(l, e);
            for(int jBar = j - l + 1; jBar <= j; jBar++) {
                product += getAlignmentProb(i, jBar, destLength, sourceSentence.length) + getTranslationProb(sourceSentence[jBar-1], e);
            }
            
            QElement bestEPrime = new QElement(Double.NEGATIVE_INFINITY, c, j, e, 0, null);
            
            WordsAndProbabilities [] wapArray = getPrevWordsAndProbs(e, i);
            for(WordsAndProbabilities wap : wapArray) {
	            String [] loopWords = wap.getWords();
	            double [] loopProbs = wap.getProbs();
	            for(int index = 0; index < loopWords.length; index++) {
	                String ePrime = loopWords[index];
	                double bigramProb = mixLMProbs(loopProbs[index], unigramProb);
	                
	                QElement bestJPrime = new QElement(Double.NEGATIVE_INFINITY, c, j, e, 0, null);
	                
	                //for loop that produces best j prime
	                Pair<Integer, Integer> jPrimeRange = getJRange(c - l, sourceSentence, i - 1, destLength);
	                int jPrimeStart = jPrimeRange.getLeft();
	                int jPrimeEnd = jPrimeRange.getRight();

	                for(int jPrime = jPrimeStart; jPrime <= jPrimeEnd; jPrime++) {
	                    QElement prev = previousLevel.getElement(c - l, jPrime, ePrime);
	                    int newCoverage = coverageToBitField(j, l);
	                    if ((newCoverage & prev.getCoverage()) == 0) {
	                        //the coverage DOES NOT OVERLAP
	                    	double newJPrimeProb = prev.getProb();
	                    	int newJPrimeCoverage = prev.getCoverage() + newCoverage;
	                    	if(i == destLength) {
	                    		newJPrimeProb += getSentenceEndProb(e, sourceSentence, newJPrimeCoverage);
	                    	}
	                        if (newJPrimeProb > bestJPrime.getProb()) {
	                            bestJPrime = new QElement(newJPrimeProb, c, j, e, newJPrimeCoverage, prev);
	                        }
	                    } 
	                }
	                
	                double newEPrimeProb = bigramProb + bestJPrime.getProb();
	                if (newEPrimeProb > bestEPrime.getProb()) {
	                    bestEPrime = new QElement(newEPrimeProb, c, j, e, bestJPrime.getCoverage(), bestJPrime.getPrevious());
	                }
	            }
            }
            
            double newLProb = product + bestEPrime.getProb();
            if (newLProb > bestL.getProb()) {
                bestL = new QElement(newLProb, c, j, e, bestEPrime.getCoverage(), bestEPrime.getPrevious());
            }
        }

        return bestL;
    }

    protected QElement computeNoJNonSkipQElement(int c, int i, String e, QLevel previousLevel, String [] sourceSentence, int destLength) {        
        double unigramProb = getUnigramProb(e);
        QElement bestJL = new QElement(Double.NEGATIVE_INFINITY, c, -1, e, 0, null);

        for(int j = 1; j <= sourceSentence.length; j++) {
	        //for loop that produces best l
	        int maxL = getFertilityMax(c, i, j, e, sourceSentence, destLength);
	        for(int l = 1; l <= maxL; l++) {
	            double product = getFertilityProb(l, e);
	            for(int jBar = j - l + 1; jBar <= j; jBar++) {
	                product += getAlignmentProb(i, jBar, destLength, sourceSentence.length) + getTranslationProb(sourceSentence[jBar-1], e);
	            }
	            
	            QElement bestEPrime = new QElement(Double.NEGATIVE_INFINITY, c, -1, e, 0, null);
	            
	            WordsAndProbabilities [] wapArray = getPrevWordsAndProbs(e, i);
	            for(WordsAndProbabilities wap : wapArray) {
		            String [] loopWords = wap.getWords();
		            double [] loopProbs = wap.getProbs();
		            for(int index = 0; index < loopWords.length; index++) {
		                String ePrime = loopWords[index];
		                double bigramProb = mixLMProbs(loopProbs[index], unigramProb);
		                
		                QElement prev = previousLevel.getElement(c - l, -1, ePrime);
		                int newCoverage = coverageToBitField(j, l);
		                if ((newCoverage & prev.getCoverage()) == 0) {
		                	double newEPrimeProb = bigramProb + prev.getProb();
		                	int newEPrimeCoverage = prev.getCoverage() + newCoverage;
		                	if(i == destLength) {
		                		// include end-of-sentence prob
		                		newEPrimeProb += getSentenceEndProb(e, sourceSentence, newEPrimeCoverage);
		                	}
		                	if (newEPrimeProb > bestEPrime.getProb()) {
		                		bestEPrime = new QElement(newEPrimeProb, c, -1, e, newEPrimeCoverage, prev);
		                	}
		                }
		            }
	            }
	            
	            double newJLProb = product + bestEPrime.getProb();
	            if (newJLProb > bestJL.getProb()) {
	                bestJL = new QElement(newJLProb, c, -1, e, bestEPrime.getCoverage(), bestEPrime.getPrevious());
	            }
	        }
        }
        
        return bestJL;
    }

    /**
     * Returns the log base ten of a double.

     @param value the number whose log-base-ten will be returned
     @return the log base ten of the number
     */

    protected static double logTen(double value) {
    	return Math.log(value) / Math.log(10);
    }

    /**
     * Prints the provided message if the class has been set to
     * print verbose messages.

     @param message the string to be printed
     */
    
    protected void printVerbose(String message) {
	if (verbose)
	    System.err.println(message);
    }

    /**
     * Performs a test of the decoder using a simple, hard-coded
     * testing corpus and set of probabilities.

     @param args the command-line arguments, which are ignored
     */

    public static void main(String[] args) {
	DPDecodeTester.test();
    }
}

/**
 * Allows for sorting comparisons between pair objects that contain
 * a sentence and its probability.
 */

class SentenceComparator implements Comparator<Pair<String, Double>> {

    /**
     * Returns the comparison value for two pair objects, such that the
     * objects will be sorted in descending order of probability.

     @param o1 the first pair object
     @param o2 the second pair object
     @return the comparison value
     */

	//NOTE: this sorts maximum-to-minimum!
	public int compare(Pair<String, Double> o1, Pair<String, Double> o2) {
		if(o1.getRight().doubleValue() < o2.getRight().doubleValue())
			return 1;
		else if(o1.getRight().doubleValue() > o2.getRight().doubleValue())
			return -1;
		else
			return 0;
	}

    /**
     * Says whether a comparator object equals another object.

     @param obj the other object
     @return whether this object equals the other
     */

	public boolean equals(Object obj) {
		return this == obj;
	}
}

/**
 * Stores the information for a position of the dynamic-programming
 * table. The object is read-only after construction.
 */

class QElement implements Comparable<QElement> {
    private double prob;
    private String currentWord;
    private QElement previous;
    private int coveredCount, lastCoveredIndex;
    private int coverageBitField;

    /**
     * Initializes the contents of the table entry.

     @param p the probability value for this table entry.
     @param c the number of positions in the source sentence that
     * have been covered by this candidate sentence
     @param j the last position covered in the source sentence
     @param e the last word of the candidate sentence
     @param bitField the bitfield encoding which positions in the
     * source sentence have been covered so far
     @param prev the dynamic-programming table entry at the previous
     * level of the table corresponding to the previous word in the
     * candidate translation
     */
    
    public QElement(double p, int c, int j, String e, int bitField, QElement prev) {
        prob = p;
        coveredCount = c;
        lastCoveredIndex = j;
        currentWord = e;
        coverageBitField = bitField;
        previous = prev;
    }

    /**
     * Returns the probability of this candidate sentence.

     @return the probability of this candidate sentence
     */

    public double getProb() {
        return prob;
    }

    /**
     * Returns the last word of this candidate sentence.

     @return the last word of this candidate sentence
     */

    public String getCurrentWord() {
        return currentWord;
    }

    /**
     * Returns the table entry of the candidate translation ending
     * in the second-to-last word of this candidate translation.

     @return the table entry of the candidate translation ending
     * in the second-to-last word of this candidate translation
     */

    public QElement getPrevious() {
        return previous;
    }

    /**
     * Returns the number of source-sentence positions covered
     * by this candidate translation.

     @return the number of source-sentence positions covered
     * by this candidate translation
     */

    public int getCoveredCount() {
        return coveredCount;
    }

    /**
     * Returns the index of the last position covered in the source sentence.

     @return the index of the last position covered in the source sentence
     */

    public int getLastCoveredIndex() {
        return lastCoveredIndex;
    }

    /**
     * Returns the number of source-sentence positions covered
     * by this candidate translation.

     @return the number of source-sentence positions covered
     * by this candidate translation
     */

    public int getC() {
        return coveredCount;
    }

    /**
     * Returns the index of the last position covered in the source sentence.

     @return the index of the last position covered in the source sentence
     */

    public int getJ() {
        return lastCoveredIndex;
    }

    /**
     * Returns the bitfield encoding which positions in the source sentence
     * have been covered.

     @return the bitfield encoding which positions in the source sentence
     * have been covered
     */

    public int getCoverage() {
        return coverageBitField;
    }

    /**
     * Returns an array of the indices that are listed as being
     * covered by the given coverage bitfield, in ascending order
     * of index.

     @param coverage the coverage bit field
     @return the array of covered indices
     */

    public static int [] coverageToIndices(int coverage) {
    	int [] allCoveredIndices = new int [32];
    	int arrayIndex = 0;

    	for(int i = 0; i < 32; i++) {
    		if((coverage & (1 << i)) != 0) {
    			allCoveredIndices[arrayIndex] = i + 1;
    			arrayIndex++;
    		}
    	}

    	int [] outArray = new int [arrayIndex];
    	for(int i = 0; i < arrayIndex; i++) {
    		outArray[i] = allCoveredIndices[i];
    	}

    	return outArray;
    }

    /**
     * Returns the candidate sentence string represented by this object.

     @param showAlignment whether to include alignment information
     * in the resulting string
     @return the candidate sentence string represented by this object
     */
    
    public String reconstructSentence(boolean showAlignment) {
        if (previous == null) {
            //return currentWord;
            return "";
        } else {
            String incrementalStr = "";
            if(showAlignment) {
            	int incrementalCoverage = coverageBitField & (~previous.getCoverage());
            	incrementalStr = Arrays.toString(coverageToIndices(incrementalCoverage)).replaceAll(" ", "");
            }
            if(previous.getCurrentWord().equals("<s>"))
                return currentWord + incrementalStr;
            return previous.reconstructSentence(showAlignment) + " " + currentWord + incrementalStr;
        }
    }

    /**
     * Returns the comparison value between this and another table
     * element for the purposes of sorting them by probability.

     @return the comparison value between this and another table
     * element
     */

	public int compareTo(QElement other) {
		double otherProb = other.getProb();
		if(prob < otherProb)
			return -1;
		else if(prob > otherProb)
			return 1;
		else
			return 0;
	}
}

/**
 * A class to allow efficient storage of QElement objects in a hashtable.
 */

class QLevelHashTuple {
    private int c;
    private int j;
    private String e;
    private int hashVal;

    /**
     * Constructs the object, using its arguments to generate a hash code.

     @param cPass the value of the c part of the index tuple for the dynamic programming table
     @param jPass the value of the j part of the index tuple for the dynamic programming table
     @param ePass the value of the e part of the index tuple for the dynamic programming table
     */

    public QLevelHashTuple(int cPass, int jPass, String ePass) {
    	c = cPass;
    	j = jPass;
    	e = ePass;
    	hashVal = c ^ j ^ e.hashCode();
    }

    /**
     * Returns the value stored in c.

     @return the integer value stored in c
     */
    public int getC() { return c; }

    /**
     * Returns the value stored in j.

     @return the integer value stored in j
     */

    public int getJ() { return j; }

    /**
     * Returns the value stored in e.

     @return the string value stored in e
     */

    public String getE() { return e; }

    /**
     * Returns a hash code for the object for the purposes of storage in
     * the dynamic programming table.

     @return the integer hash value
    */

    @Override
    public int hashCode() {
        return hashVal;
    }

    /**
     * Determines if this object is equal to another of the same type.

     @param o the other object
     @return whether the objects are equal
    */

    @Override
    public boolean equals(Object o) {
        if (o == null) return false;
        if (!(o instanceof QLevelHashTuple)) return false;
        QLevelHashTuple other = (QLevelHashTuple) o;
        return (c == other.getC()) && (j == other.getJ()) && e.equals(other.getE());
    }
}

/**
 * A single level in the dynamic programming table, where a level is defined
 * by a fixed value of i. The value of i is equal to the current position in
 * the decoded sentence that is being formed.
 */

class QLevel {
    private Hashtable<QLevelHashTuple, QElement> elementsTable;
    private ArrayList<QElement> elementsList;
    private int level;
   
    /**
     * Constructs and initializes the object for a specified level.
     
     @param l the current level
    */

    public QLevel(int l) {
        level = l;
        elementsTable = new Hashtable<QLevelHashTuple, QElement>();
        elementsList = new ArrayList<QElement>();
    }

    /**
     * Returns the number of elements in the level.

     @return the number of elements in the level
     */

    public int size() {
    	return elementsList.size();
    }
    
    /**
     * Adds a QElement to the table. Elements with zero probability (negative
     * infinity log probability) are not added to the table to save space
     * and time.

     @param the element to add
     */

    public void addElement(QElement element) {
        if(element.getProb() > Double.NEGATIVE_INFINITY)
        {
            elementsTable.put(new QLevelHashTuple(element.getC(), element.getJ(), element.getCurrentWord()), element);
            elementsList.add(element);
        }
    }
    
    /**
     * Determine if an element with the specified values exists in the table.
     
     @param c the c value of the element
     @param j the j value of the element
     @param e the e value of the element
     @return whether a matching element exists in the table
     */

    public boolean hasElement(int c, int j, String e) {
        return elementsTable.containsKey(new QLevelHashTuple(c, j, e));
    }
    
    /**
     * Return a matching QElement if it exists in the table. If there is no
     * matching element, return a dummy QElelement with 0 probability
     * (negative infinity log probability).

     @param c the c value of the element
     @param j the j value of the element
     @param e the e value of the element
     @return the matching element, or a dummy if no match
     */

    public QElement getElement(int c, int j, String e) {
        if(hasElement(c, j, e)) {
            return elementsTable.get(new QLevelHashTuple(c, j, e));
        } else
	     //dummy element; maximization in algorithms should skip this
            return new QElement(Double.NEGATIVE_INFINITY, c, j, e, 0, null); 
    }

    /**
     * Returns all unique pairs of c,j values at the current level. This is
     * intended to reduce the search space at the next level. NOTE: this
     * optimization may be obsolete, no longer providing any benefit.

     @return an arraylist of pairs of c,j values
     */
    
    public ArrayList<Pair<Integer, Integer>> getCJPairs() {
        ArrayList<Pair<Integer, Integer>> cjList = new ArrayList<Pair<Integer, Integer>>();
        HashSet<Pair<Integer, Integer>> cjSet = new HashSet<Pair<Integer, Integer>>();
        for (QElement element : elementsList) {
            Pair<Integer, Integer> currentPair = new Pair<Integer, Integer>(new Integer(element.getC()), new Integer(element.getJ()));
            if (! cjSet.contains(currentPair)) {
                cjSet.add(currentPair);
                cjList.add(currentPair);
            }
        }
        return cjList;
    }

    /**
     * Returns the element with the highest probability that covers a
     * specified amount of the source sentence. This coverage constraint
     * is used during decoding to ensure that candidates achieve full
     * coverage.

     @param numCovered the desired coverage (c)
     @return the highest probability element, or a dummy with zero probability
     */
    
    public QElement getMaxElement(int numCovered) {
        QElement maxElement = new QElement(Double.NEGATIVE_INFINITY, 0, -1, "<s>", 0, null);
        for(QElement element : elementsList) {
            if(element.getProb() > maxElement.getProb() && element.getCoveredCount() == numCovered) {
                maxElement = element;
            }
        }
        return maxElement;
    }

    /**
     * Returns the elements with the highest probabilies.

     @param maxResults the maximum number of results
     @return an array of the highest probability elements
    */

    public QElement [] getMaxElements(int maxResults) {
    	int size = elementsList.size();
    	PriorityQueue<QElement> maxElements;

        /*
        for(QElement element : elementsList) {
        	//UBERHACK!
        	if(true || element.getCoveredCount() == numCovered) {
        		size++;
        	}
        }
        */

        if(size == 0) {
        	return new QElement[0];
        }

    	maxElements = new PriorityQueue<QElement>(size, Collections.reverseOrder());
        for(QElement element : elementsList) {
        	maxElements.add(element);
        }
        int arraySize = size;
        if(arraySize > maxResults)
        	arraySize = maxResults;
        QElement [] maxElementsArray = new QElement[arraySize];
        for(int i = 0; i < arraySize; i++) {
        	maxElementsArray[i] = maxElements.poll();
        }
        return maxElementsArray;
    }
    
}
/**
 * A class to represent a pair of objects, as java has no built in tuple. This
 * class was taken from the following stack overflow question:
 * http://stackoverflow.com/questions/521171/a-java-collection-of-value-pairs-tuples
 */
class Pair<L,R> {

    private final L left;
    private final R right;

    /**
     * Constructs the class given two objects.

     @param left the first object
     @param right the second object
     */

    public Pair(L left, R right) {
        this.left = left;
        this.right = right;
    }

    /**
     * Returns the first (left) object.

     @return the first object
     */

    public L getLeft() { return left; }

    /**
     * Returns the second (right) object.

     @return the second object
     */

    public R getRight() { return right; }

    /**
     * Returns a hash code for the pair of objects.

     @return a hash code
     */

    @Override
    public int hashCode() { return left.hashCode() ^ right.hashCode(); }


    /**
     * Determines if this object is equal to another object of the same type.

     @param o the other object
     @return whether the objects are equal
     */

    @Override
    public boolean equals(Object o) {
        if (o == null) return false;
        if (!(o instanceof Pair)) return false;
        Pair pairo = (Pair) o;
        return this.left.equals(pairo.getLeft()) && this.right.equals(pairo.getRight());
    }
}

/**
 * A class that implements a CorpusAccessor, overriding most of the functionality
 * for testing purposes. This allows us to more easily test the decoder on a 
 * constructed corpus without having to actually create it in a file and train
 * on it.
 */

class TestingCorpusAccessor implements CorpusAccessor {
    
    public int getSentenceLength(int index, boolean isDestLanguage) {
        return 0;
    }

    public String getWord(int sentenceIndex, boolean isDestLanguage, int wordIndex) {
        return "GET TO THE CHOPPER!";
    }
    
    public String[] getSentence(int index, boolean isDestLanguage) {
        return null;
    }
    
    public String[][] getSentencePair(int index) {
        return null;
    }
    
    public int getSize() {
        return 0;
    }
    
    public String[] getWordList(boolean isDestLanguage) {
        String [] wordList = new String[5];
        for(int i = 0; i < 5; i++) {
            wordList[i] = "e" + Integer.toString(i);
        }
        return wordList;
    }
}

class TestingLanguageModel implements LanguageModel {
    public double queryNgram(String ngram) {
    	return 0.0; //full probability
    }
}

/**
 * An implementation of DPDecoder with the ability to tweak important
 * functionality for testing purposes.

 @see DPDecoder
 */

class DPDecodeTester extends DPDecoder {
    public DPDecodeTester(ParameterAccessor p, CorpusAccessor c, LanguageModel lm, SortedProbabilityAccessor ntp) {
	super(p, c, lm, ntp);

	useStartingBigrams = false;
	useEndingBigrams = false;
	restrictCoverage = false;
	restrictLastCovered = false;
	restrictFertility = false;
	ignoreLastCovered = false;
    }

    protected void preprocessSentence(String sourceSentence) {
    	preprocessedSourceSentence = sourceSentence;
    	outerLoopWords = uniqueDestWords;
    }

    protected double getUnigramProb(String unigram) {
    	return 0.0; //full log prob
    }

    protected double mixLMProbs(double bigramProb, double unigramProb) {
    	return bigramProb;
    }

    protected WordsAndProbabilities [] getPrevWordsAndProbs(String e, int i) {
	WordsAndProbabilities [] wapArray = new WordsAndProbabilities[1];
	wapArray[0] = nextToPrevs.getInnerData(e);
	return wapArray;
    }

    protected double getTranslationProb(String f, String e) {
        assert(f.length() == 2 && e.length() == 2);
        int firstInt = Integer.parseInt(e.substring(1, 2));
        int secondInt = Integer.parseInt(f.substring(1, 2));
        if(firstInt == secondInt)
            return 0.0;
        return Double.NEGATIVE_INFINITY;
    }

    protected double getFertilityProb(int l, String e) {
    	return 0.0;
    }

    protected double getAlignmentProb(int i, int j, int I, int J) {
    	return 0.0;
    }

    public static void test() {
        TestingCorpusAccessor corpus = new TestingCorpusAccessor();
        //TestingParameterAccessor params = new TestingParameterAccessor();
        //TestingLanguageModel languageModel = new TestingLanguageModel();

        //populate testing nextToPrevs
        SortedProbabilityAccessor nextToPrevs = new SortedProbabilityAccessor();
        //e1 -> e1 (0.5) or <s> (0.2); ei -> ei (0.5), e[i-1] (0.5), <s> (0.2)
        String nextWord = "e1";
        String [] prevWords = new String[2];
        double [] prevProbs = new double[2];
        prevWords[0] = "<s>";         prevWords[1] = "e1";
        prevProbs[0] = logTen(0.2);   prevProbs[1] = logTen(0.5);
        nextToPrevs.populateOuterWord(nextWord, prevWords, prevProbs);
        for(int i = 2; i <= 5; i++) {
        	String iStr = Integer.toString(i);
        	nextWord = "e" + iStr;
        	prevWords = new String[3];
        	prevProbs = new double[3];

        	prevWords[0] = "<s>";
        	prevProbs[0] = logTen(0.2);

        	prevWords[1] = "e" + Integer.toString(i - 1);
        	prevProbs[1] = logTen(0.5);

        	prevWords[2] = nextWord;
        	prevProbs[2] = logTen(0.5);

        	nextToPrevs.populateOuterWord(nextWord, prevWords, prevProbs);
        }

        //construct decoder
        DPDecoder dec = new DPDecodeTester(null, corpus, new TestingLanguageModel(), nextToPrevs);

        System.out.println("-------Testing testing corpus accessor-------");
        String [] udw = corpus.getWordList(true);
        System.out.println("Start word list");
        for(String w : udw)
            System.out.println(w);
        System.out.println("End word list");
        
        //Test of base case response in skip and nonskip
        String sourceSentence;
        String [] splitSourceSentence;
        String [] emptySentence = {""};
        QElement element;
        
        QLevel previousLevel = new QLevel(0);
        QElement baseCaseElement = new QElement(0.0, 0, -1, "<s>", 0, null);
        previousLevel.addElement(baseCaseElement);
        
        System.out.println("-------Testing base-case compute skip elements-------");
        element = dec.computeSkipQElement(0, 1, -1, "e1", previousLevel, emptySentence, 4);
        System.out.println("expecting 1/5 from skip(0, 1, -1, e1, base case), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeSkipQElement(0, 1, -1, "e2", previousLevel, emptySentence, 4);
        System.out.println("expecting 1/5 from skip(0, 1, -1, e2, base case), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeSkipQElement(0, 1, -1, "e3", previousLevel, emptySentence, 4);
        System.out.println("expecting 1/5 from skip(0, 1, -1, e3, base case), got " + Double.toString(Math.pow(10, element.getProb())));
        //examine other members
        System.out.println("expecting current word e3 from element, got " + element.getCurrentWord());
        System.out.println("expecting base-case from element (<s>), got " + element.getPrevious().getCurrentWord());
        System.out.println("expecting covered count 0 from element, got " + Integer.toString(element.getCoveredCount()));
        System.out.println("expecting last covered index -1 from element, got " + Integer.toString(element.getLastCoveredIndex()));
        System.out.println("expecting coverage 0 from element, got " + Integer.toString(element.getCoverage()));
        
        //(int c, int i, int j, String e, QLevel previousLevel, String [] sourceSentence, int destLength)
        
        System.out.println("-------Testing base-case compute non-skip elements-------");
        sourceSentence = "s1 s2 s3 s4";
        splitSourceSentence = sourceSentence.split("\\s+");
        System.out.println("For source sentence \"" + sourceSentence + "\":");
        element = dec.computeNonSkipQElement(1, 1, 1, "e1", previousLevel, splitSourceSentence, 4);
        System.out.println("1/5: nonskip(1, 1, 1, e1, base case, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeNonSkipQElement(1, 1, 1, "e0", previousLevel, splitSourceSentence, 4);
        System.out.println("0: nonskip(1, 1, 1, e0, base case, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeNonSkipQElement(1, 1, 1, "e3", previousLevel, splitSourceSentence, 4);
        System.out.println("0: nonskip(1, 1, 1, e3, base case, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeNonSkipQElement(1, 1, 3, "e3", previousLevel, splitSourceSentence, 4);
        System.out.println("1/5: nonskip(1, 1, 3, e3, base case, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        
        sourceSentence = "s2 s2 s3 s4";
        splitSourceSentence = sourceSentence.split("\\s+");
        System.out.println("For source sentence \"" + sourceSentence + "\":");
        element = dec.computeNonSkipQElement(2, 1, 1, "e2", previousLevel, splitSourceSentence, 4);
        System.out.println("0: nonskip(2, 1, 1, e2, base case, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeNonSkipQElement(2, 1, 3, "e2", previousLevel, splitSourceSentence, 4);
        System.out.println("0: nonskip(2, 1, 3, e2, base case, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeNonSkipQElement(2, 1, 2, "e2", previousLevel, splitSourceSentence, 4);
        System.out.println("1/5: nonskip(2, 1, 2, e2, base case, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        System.out.println("expecting current word e2 from element, got " + element.getCurrentWord());
        System.out.println("expecting base-case from element (<s>), got " + element.getPrevious().getCurrentWord());
        System.out.println("expecting covered count 2 from element, got " + Integer.toString(element.getCoveredCount()));
        System.out.println("expecting last covered index 2 from element, got " + Integer.toString(element.getLastCoveredIndex()));
        System.out.println("expecting coverage 3 from element, got " + Integer.toString(element.getCoverage()));
        
        System.out.println("-------Testing first non-base-case compute skip elements-------");
        
        previousLevel = new QLevel(1);
        previousLevel.addElement(new QElement(logTen(0.2), 0, -1, "e1", 0, baseCaseElement));
        previousLevel.addElement(new QElement(logTen(0.2), 0, -1, "e2", 0, baseCaseElement));
        
        
        element = dec.computeSkipQElement(0, 2, -1, "e3", previousLevel, emptySentence, 4);
        System.out.println("expecting 0.1 from skip(0, 2, -1, e3, prevLevel), got " + Double.toString(Math.pow(10, element.getProb())));
        System.out.println("expecting e2 from prev's word, got " + element.getPrevious().getCurrentWord());
        element = dec.computeSkipQElement(0, 2, -1, "e2", previousLevel, emptySentence, 4);
        System.out.println("expecting 0.1 from skip(0, 2, -1, e2, prevLevel), got " + Double.toString(Math.pow(10, element.getProb())));
        System.out.println("expecting e1 from prev's word, got " + element.getPrevious().getCurrentWord());
        element = dec.computeSkipQElement(0, 2, -1, "e4", previousLevel, emptySentence, 4);
        System.out.println("expecting 0 from skip(0, 2, -1, e4, prevLevel), got " + Double.toString(Math.pow(10, element.getProb())));
        
        System.out.println("-------Testing first non-base-case compute non-skip elements-------");
        previousLevel = new QLevel(1);
        previousLevel.addElement(new QElement(logTen(0.2), 1, 3, "e3", 4, baseCaseElement));
        
        element = dec.computeNonSkipQElement(2, 2, 3, "e3", previousLevel, splitSourceSentence, 4);
        System.out.println("0: nonskip(2, 2, 3, e3, prevLevel, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        element = dec.computeNonSkipQElement(2, 2, 2, "e2", previousLevel, splitSourceSentence, 4);
        System.out.println("0: nonskip(2, 2, 2, e2, prevLevel, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        
        previousLevel.addElement(new QElement(logTen(0.2), 1, 1, "e2", 1, baseCaseElement));
        
        element = dec.computeNonSkipQElement(2, 2, 3, "e3", previousLevel, splitSourceSentence, 4);
        System.out.println("0.1: nonskip(2, 2, 3, e3, prevLevel, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        System.out.println("expecting e2 from prev's word, got " + element.getPrevious().getCurrentWord());
        element = dec.computeNonSkipQElement(2, 2, 2, "e2", previousLevel, splitSourceSentence, 4);
        System.out.println("0.1: nonskip(2, 2, 2, e2, prevLevel, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        System.out.println("expecting e2 from prev's word, got " + element.getPrevious().getCurrentWord());
        element = dec.computeNonSkipQElement(3, 2, 2, "e2", previousLevel, splitSourceSentence, 4);
        System.out.println("0: nonskip(3, 2, 2, e2, prevLevel, sentence, 4), got " + Double.toString(Math.pow(10, element.getProb())));
        
        System.out.println("-------Testing full decodeToLength-------");
        
        Pair<String, Double>result;
        
        result = dec.decodeToLength(sourceSentence, 4);
        System.out.println("expecting e2 e2 e3 e4 by translating " + sourceSentence + " to length-4 " + result.getLeft());
        System.out.println("with probability " + Double.toString(Math.pow(10, result.getRight().doubleValue())));
        result = dec.decodeToLength(sourceSentence, 3);
        System.out.println("expecting e2 e3 e4 by translating " + sourceSentence + " to length-3 " + result.getLeft());
        System.out.println("with probability " + Double.toString(Math.pow(10, result.getRight().doubleValue())));
        result = dec.decodeToLength(sourceSentence, 5);
        System.out.println("translating " + sourceSentence + " to length-5 " + result.getLeft());
        System.out.println("with probability " + Double.toString(Math.pow(10, result.getRight().doubleValue())));
        
        sourceSentence = "s2 s2 s3 s3 s4 s4";
        splitSourceSentence = sourceSentence.split("\\s+");
        
        result = dec.decodeToLength(sourceSentence, 3);
        System.out.println("translating " + sourceSentence + " to length-3 " + result.getLeft());
        System.out.println("with probability " + Double.toString(Math.pow(10, result.getRight().doubleValue())));
        result = dec.decodeToLength(sourceSentence, 4);
        System.out.println("translating " + sourceSentence + " to length-4 " + result.getLeft());
        System.out.println("with probability " + Double.toString(Math.pow(10, result.getRight().doubleValue())));
        result = dec.decodeToLength(sourceSentence, 5);
        System.out.println("translating " + sourceSentence + " to length-5 " + result.getLeft());
        System.out.println("with probability " + Double.toString(Math.pow(10, result.getRight().doubleValue())));
        result = dec.decodeToLength(sourceSentence, 2);
        System.out.println("SHOULD FAIL translating " + sourceSentence + " to length-2 " + result.getLeft());
        System.out.println("with probability " + Double.toString(Math.pow(10, result.getRight().doubleValue())));	
    }
}
