import java.io.*;
import java.util.*;

public class DecodeRunner {

	public static String expandUTF8(String str) {
		// expand unicode characters to utf-8
		// multi-character representations

		// to solve the issue that a sentence received by
		// Java as a command-line argument has non-ASCII
		// utf-8 characters converted to single Unicode
		// characters in the string, whereas all the rest
		// of the code works with the multi-byte
		// representations of such characters

		try {
			byte [] utf8bytes = str.getBytes("UTF-8");
			String utf8str = new String(utf8bytes);
			//System.out.println("utf-8-ified: " + utf8str);
			return utf8str;
		} catch (UnsupportedEncodingException e) {
			System.out.println("UTF-8 CONVERSION FAILED!");
			return "";
		}
	}

    public static void main(String [] args) {
	if (args.length != 5 || (args.length == 4 && ! args[3].equals("-i"))) {
	    System.err.println("Usage: DecodeRunner sourceLang targetLang [dp1|dp2|dp3|as] -f filename\n"
			      +"                                                           -s sentence\n"
			      +"                                                           -i");
	    return;
	}

	String source = args[0];
	String target = args[1];
	String decoder_model = args[2];

	LanguagePair lp = new LanguagePair(source, target);
	if (! lp.validPair) {
	    System.err.println("Invalid or unsupported language pair: " + source + ", " + target);
	    return;
	}
	if (lp.sourceTextFile == null || lp.destTextFile == null) {
	    System.err.println("Missing corpus files for language pair: "+ source + ", " + target);
	    return;
	}
	if (lp.srilmFile == null) {
	    System.err.println("Missing (SRILM) language file for language pair: "+ source + ", " + target);
	    return;
	}
	if (lp.sbpFile == null) {
	    System.err.println("Missing sorted probability accessor for language pair: "+ source + ", " + target);
	    return;
	}
	System.err.println("Loaded language pair:" + source + ", " + target);


	System.err.println("Loading parameter files...");

        CorpusFromTextPair corpus = new CorpusFromTextPair(lp.sourceTextFile, lp.destTextFile);
        BigramsFromSrilmFile bfsf = new BigramsFromSrilmFile(lp.srilmFile);
        SortedProbabilityAccessor nextToPrevs = new SortedProbabilityAccessor(lp.sbpFile);
	System.err.println("Parameters loaded");
	System.err.println("Setting up decoder");
	Decoder dec;
	if (decoder_model.equals("dp1")) {
	    if (lp.model1Params == null) {
		System.err.println("Missing model 1 translation params for language pair: "+ source + ", " + target);
		return;
	    }
	    ParametersFromTextFile params = new ParametersFromTextFile(lp.model1Params);
	    dec = new DPModel1Decoder(params, corpus, bfsf, nextToPrevs);
	} else if (decoder_model.equals("dp2")) {
	    if (lp.model2Params == null) {
		System.err.println("Missing model 2 translation params for language pair: "+ source + ", " + target);
		return;
	    }
	    ParametersFromTextFile params = new ParametersFromTextFile(lp.model2Params);
	    dec = new DPModel2Decoder(params, corpus, bfsf, nextToPrevs);
	} else if (decoder_model.equals("dp3")) {
	    if (lp.model3Params == null) {
		System.err.println("Missing model 3 translation params for language pair: "+ source + ", " + target);
		return;
	    }
	    ParametersFromTextFile params = new ParametersFromTextFile(lp.model3Params);
	    dec = new DPModel3Decoder(params, corpus, bfsf, nextToPrevs);
	} else if (decoder_model.equals("as")) {
	    if (lp.model3Params == null) {
		System.err.println("Missing model 3 translation params for language pair: "+ source + ", " + target);
		return;
	    }
	    Model3Params params = new Model3Params(lp.model3Params);
	    dec = new ASDecoder(params, corpus, bfsf, nextToPrevs);
	} else {
	    System.err.println("Invalid decoder/model: " + decoder_model);
	    System.err.println("Available decoders and models: dp1, dp2, dp3, as");
	    return;
	}
	System.err.println("Decoder set up.");

	if (args[3].equals("-s")) {
	    String input = expandUTF8(args[4]);
	    System.err.println("Decoding sentence: "+ input);
	    String result = dec.decode(input);
	    if (result == null) {
		System.err.println("No result");
	    } else {
		System.out.println(result);
	    }
	} else if (args[3].equals("-f")) {
	    String filename = args[4];
	    System.err.println("Decoding file: "+ filename);
	
	    try {
		BufferedReader in = new BufferedReader(new FileReader(filename));
		String sourceSentence;
		while ((sourceSentence = in.readLine()) != null) {
		    System.err.println("Decoding sentence: " + sourceSentence);
		    System.out.println(dec.decode(sourceSentence));
		}
		in.close();
	    } catch (IOException e) {
		System.err.println("Failed to read from: " + filename);
		    }
	} else if (args[3].equals("-i")) {
	    System.err.println("Interactive Decode:");
	    try {
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		String sourceSentence;
		while ((sourceSentence = in.readLine()) != null) {
		    System.out.println(dec.decode(sourceSentence));
		}
		in.close();
	    } catch (IOException e) {
		System.err.println("User input error");
	    }
	} else {
	    System.err.println("Invalid mode flag (expected s, f, or i): " + args[3]);
	}
    }
}
