import java.io.*;
import java.util.*;
import java.lang.Math;

class Model1Decoder {
    
    private LMClient languageModel;
    private Model1 translationModel;
    private CorpusFromTextPair corpus;
    private ParametersFromTextFile params;
    private String[] uniqueDestWords;
    private Hashtable<String, HashSet<String>> wordMap;
    
    public Model1Decoder() throws IOException{
        languageModel = new LMClient();
        translationModel = new Model1();
        languageModel.open("127.0.0.1", 50000);
        double result = languageModel.queryNgram("thank you");
        
        params = new ParametersFromTextFile();
        corpus = new CorpusFromTextPair("../corpora/textpairs/europarl_2-to-10_stripped.es-en.es", "../corpora/textpairs/europarl_2-to-10_stripped.es-en.en");
        System.out.println("Getting word list");
        uniqueDestWords = corpus.getWordList(true);
        //translationModel.initialize(params, corpus);
        //translationModel.train(params, corpus, 5);
        System.out.println("Importing parameters");
        params.importParams("../parameters/textfiles/europarl_2-to-10_stripped.es-en.model1.5iters");
        
        System.out.println("Getting word map");
        wordMap = corpus.getWordMap();
        System.out.println("Word map complete");        
    }

    public String decode(String sourceSentence) throws IOException{
    
        String[] sourceSentenceWords = sourceSentence.split(" ");
    
        int sourceLength = sourceSentenceWords.length;
        
        String destSentence = "";
        String prevDestWord = "<s>";
        
        for (int i = 0; i < sourceLength; i++) {
            
            String nextWord = this.getBestWord(prevDestWord, sourceSentenceWords[i]);
            destSentence += " "+nextWord;
            prevDestWord = nextWord;
        }
        destSentence = destSentence.substring(1);
        System.out.println("Prototype: "+destSentence);
        System.out.println("Starting hill climbing...");
        String[] result = hillClimb(destSentence, sourceSentence, 1);
        return join(result);
    }
        
    public String getBestWord(String context, String sourceWord) throws IOException{
        String bestWord = "";
        double bestProb = Double.NEGATIVE_INFINITY;
        
        for (String destWord : uniqueDestWords) {
            // NOTE: these are log-probabilities (base 10)
            double ngramProb = languageModel.queryNgram(context + " " + destWord);
            double translationProb = Math.log(params.getT(sourceWord, destWord))/Math.log(10);
            double currentProb = ngramProb + translationProb;
            if (currentProb > bestProb) {
                bestProb = currentProb;
                bestWord = destWord;
            }
        }
        if (bestWord == "") {
            return sourceWord;
        } else {
            return bestWord;
        }
    }
    
    public String[] hillClimb(String startingDestSentence, String sourceSentence, int numResults) {
		String[] bestDestWords = startingDestSentence.split("\\s+");
		String[] sourceWords = sourceSentence.split("\\s+");
        //String[][] bestDestSentences = String[][];

        double bestProb = translationProbability(bestDestWords, sourceWords);

        String[] destWords = bestDestWords.clone();
            
        double workingBestProb;
        
        for (int i=0; i < destWords.length; i++) {
            String[] workingDestWords = destWords.clone();
            //System.out.print("Hill "+Integer.toString(i)+"\r\n\r\n");            
            substitute(workingDestWords, sourceWords, i);


            
            for (int j=0; j < workingDestWords.length; j++) {
                permute(workingDestWords, sourceWords, j);
                
                substitute(workingDestWords, sourceWords, j);
            }
            /*
            //for (int j=0; j < workingDestWords.length; j++) {
            permute(workingDestWords, sourceWords, i);
            
            boolean changed = true;
            int iter = 0;
            while (changed) {
                changed = false;
                System.out.println("Iteration: "+Integer.toString(iter));
                for (int k=0; k < workingDestWords.length; k++) {
                    if (substitute(workingDestWords, sourceWords, k))
                        changed = true;
                }
                iter++;
            }
            
            permute(workingDestWords, sourceWords, i);
            //}*/
            workingBestProb = translationProbability(workingDestWords, sourceWords);
            if (workingBestProb > bestProb) {
                bestProb = workingBestProb;
                bestDestWords = workingDestWords;
            }
        }
        System.out.println("The log probability of the following sentence was: "+Double.toString(bestProb));
		return bestDestWords;
    }
    
    private boolean permute(String[] startingDestSentence, String[] sourceSentence, int position) {
        assert (position < startingDestSentence.length);
        double bestProb = translationProbability(startingDestSentence, sourceSentence);
        boolean changed = false;
        //System.out.println("permuting:    "+Arrays.toString(startingDestSentence));
        
        for (int i=position+1; i<startingDestSentence.length; i++) {
            arraySwap(startingDestSentence, position, i);
            double transProb = translationProbability(startingDestSentence, sourceSentence);
            if ( transProb > bestProb) {
                changed = true;
                bestProb = transProb;
                //System.out.println("better:       "+Arrays.toString(startingDestSentence));
                //System.out.println("with prob. : "+Double.toString(bestProb));
            } else {
                arraySwap(startingDestSentence, position, i);
            }
        }
        return changed;
    }
    
    private boolean substitute(String[] startingDestSentence, String[] sourceSentence, int position) {
        assert (position < startingDestSentence.length);
        double bestProb = translationProbability(startingDestSentence, sourceSentence);
        boolean changed = false;
        //System.out.println("substituting: "+Arrays.toString(startingDestSentence));
        
        HashSet<String> possibleTranslations = wordMap.get(sourceSentence[position]);
        if (possibleTranslations != null) {
            String lastSubstitution = startingDestSentence[position];
            for (String word : possibleTranslations) {
                startingDestSentence[position] = word;
                double transProb = translationProbability(startingDestSentence, sourceSentence);
                if ( transProb > bestProb) {
                    changed = true;
                    bestProb = transProb;
                    lastSubstitution = word;
                    //System.out.println("better:       "+Arrays.toString(startingDestSentence));
                    //System.out.println("with prob. : "+Double.toString(bestProb));
                } else {
                    startingDestSentence[position] = lastSubstitution;
                }
            }
        }
        return changed;
    }
    
    private double translationProbability(String[] destSentence, String[] sourceSentence) {
        double languageProb = 0.0;
        double translationProb = Math.log(translationModel.sentenceProbability(params, sourceSentence, destSentence));
        try {
            languageProb = languageModel.querySentence(destSentence);
        } catch (IOException e) {
            System.out.println("Weird IO stuff");
        }
        //System.out.println("language prob, translation prob: "+Double.toString(languageProb)+" || "+Double.toString(translationProb));
        return translationProb+languageProb;
    }
    
    private void arraySwap(String[] array, int pos1, int pos2) {
        String temp = array[pos1];
        array[pos1] = array[pos2];
        array[pos2] = temp;
    }
    
    private String join(String[] array) {
        String result = "";
        for (int i=0; i<array.length; i++) {
            if (i==0)
                result += array[i];
            else
                result += " " + array[i];
        }
        return result;
    }
    
    public static void main(String[] args) throws IOException {
        Model1Decoder decoder = new Model1Decoder();
        BufferedReader reader = new BufferedReader(new InputStreamReader(System.in)); 
        String input = "monkies";
        while (true) {
            System.out.print("Enter a spanish sentence to translate or ENTER to quit: ");
            input = reader.readLine();
            if (input.equals("")) break;
            System.out.println("Translating...");
            String result = decoder.decode(input);
            System.out.println(input + "\n" + "=> " +result);
            //System.out.println(Arrays.toString(decoder.hillClimb(result, input, 1)));
        }
    }
}