import java.util.*;
import java.io.*;

class BigramsFromSrilmFile implements LanguageModel {
    private Hashtable<String, Double> bigramProbs;
    private Hashtable<String, Double> unigramProbs;
    private Hashtable<String, BigramArrayLists> prevToNexts;
    private Hashtable<String, BigramArrayLists> nextToPrevs;

    public BigramsFromSrilmFile() {
	bigramProbs = new Hashtable<String, Double>();
	unigramProbs = new Hashtable<String, Double>();
	prevToNexts = new Hashtable<String, BigramArrayLists>();
	nextToPrevs = new Hashtable<String, BigramArrayLists>();
    }

    public BigramsFromSrilmFile(String filename) {
	bigramProbs = new Hashtable<String, Double>();
	unigramProbs = new Hashtable<String, Double>();
	prevToNexts = new Hashtable<String, BigramArrayLists>();
	nextToPrevs = new Hashtable<String, BigramArrayLists>();
	importSrilmFile(filename);
    }

    public void importSrilmFile(String filename) {
	BufferedReader inputStream = null;
	try {
	    inputStream = new BufferedReader(new FileReader(filename));
	    String line;
	    // 1: in 1-grams, 2: in 2-grams, 0: other
	    int state = 0; 
	    while ((line = inputStream.readLine()) != null) {
		if (line.length() < 1)
		    continue;

		if (line.charAt(0) == '\\') {
		    if (line.equals("\\1-grams:")) {
			state = 1;
		    } else if (line.equals("\\2-grams:")) {
			state = 2;
		    } else {
			state = 0;
		    }
		    continue;		    
		}

		String[] splitLine = line.split("\t");

		switch(state) {
		case 1:
		    //assert(splitLine.length == 3);
		    double unigramProb = Double.NEGATIVE_INFINITY;
		    try {
			unigramProb = Double.parseDouble(splitLine[0]);
		    } catch (NumberFormatException e) {
			System.out.println("Warning: Line contained non-numerical values: " + splitLine[1]);
		    }
		    unigramProbs.put(splitLine[1], unigramProb);
		    break;

		case 2:
		    //assert(splitLine.length == 2);
		    double bigramProb = Double.NEGATIVE_INFINITY;
		    try {
			bigramProb = Double.parseDouble(splitLine[0]);
		    } catch (NumberFormatException e) {
			System.out.println("Warning: Line contained non-numerical values: " + splitLine[1]);
		    }
		    String[] splitBigram = splitLine[1].split(" ");
		    assert(splitBigram.length == 2);

		    //System.out.println("Read bigram with probability " + Double.toString(bigramProb) + " for " + splitBigram[0] + ", " + splitBigram[1]);

		    // update hash tables
		    bigramProbs.put(splitLine[1], bigramProb);
		    updateTable(prevToNexts, splitBigram[0], splitBigram[1], bigramProb);
		    updateTable(nextToPrevs, splitBigram[1], splitBigram[0], bigramProb);
		    break;

		default:
		    break;
		}
	    }
	    inputStream.close();
	} catch (IOException e) {
	    System.out.println("Failed to import filename: " + filename);
	}
    }

    public void updateTable(Hashtable<String, BigramArrayLists> table, String outerWord, String innerWord, double bigramProb) {
	if (! table.containsKey(outerWord)) {
	    table.put(outerWord, new BigramArrayLists());
	}
	table.get(outerWord).addWordAndProb(innerWord, bigramProb);
    }

    public double queryNgram(String ngram) {
    	if(bigramProbs.containsKey(ngram)) {
    		return bigramProbs.get(ngram);
    	} else if(unigramProbs.containsKey(ngram)) {
    		return unigramProbs.get(ngram);
    	} else {
    		return Double.NEGATIVE_INFINITY;
    	}
    }

    public SortedProbabilityAccessor generateSortedProbabilities(boolean isNextToPrev) {
    	Hashtable<String, BigramArrayLists> table = prevToNexts;
    	if (isNextToPrev)
    		table = nextToPrevs;
    	
    	SortedProbabilityAccessor result = new SortedProbabilityAccessor();

        // loop over unique "outer" words
        Enumeration en = table.keys();
        while (en.hasMoreElements()) {
        	String key = (String) en.nextElement();
        	BigramArrayLists value = table.get(key);

        	// sort "inner" words/probs in descending order of probability
        	PriorityQueue<NamedProbability> queue;
        	int size = value.getWords().size();
        	queue = new PriorityQueue<NamedProbability>(size, Collections.reverseOrder());
        	for(int i = 0; i < size; i++) {
        		queue.add(new NamedProbability(value.getWords().get(i), value.getProbs().get(i)));
        	}

        	String[] innerWords = new String[size];
        	double[] innerProbs = new double[size];

        	for(int i = 0; i < size; i++) {
        		NamedProbability nextPair = queue.poll();
        		innerWords[i] = nextPair.getName();
        		innerProbs[i] = nextPair.getValue();
        	}

        	//put this data for the outer word into result
        	result.populateOuterWord(key, innerWords, innerProbs);
        }
        return result;
    }

    public void printDiagnostics() {
    	//very temporary
    	BigramArrayLists nexts;
    	int size;
    	nexts = prevToNexts.get("car");
    	System.out.println("? length of ArrayLists is " + nexts.getWords().size());
    	size = nexts.getWords().size();
    	for(int i = 0; i < size; i++) {
    		System.out.println(nexts.getWords().get(i) + ", " + Double.toString(nexts.getProbs().get(i)));
    	}
    	BigramArrayLists prevs;
    	prevs = nextToPrevs.get("naturally");
    	System.out.println("? length of ArrayLists is " + prevs.getWords().size());
    	size = prevs.getWords().size();
    	for(int i = 0; i < size; i++) {
    		System.out.println(prevs.getWords().get(i) + ", " + Double.toString(prevs.getProbs().get(i)));
    	}
    }

    public static void main(String [] args) {
    	
    	if (args.length != 3) {
	    System.out.println("usage: java BigramsFromSrilmFile srilmLMFile sbpOutFile [ntp|ptn]");
	    return;
	}
	
	String srilmFile = args[0];
	String sbpFile = args[1];

	boolean isNTP = true;
	if (args[2].equals("ntp"))
	    isNTP = true;
	else if (args[2].equals("ptn"))
	    isNTP = false;
	else {
	    System.out.println("Unrecognized argument (expecting ntp or ptn): " + args[2]);
	    return;
	}
	    

	BigramsFromSrilmFile bigrams = new BigramsFromSrilmFile(srilmFile);
    	SortedProbabilityAccessor spa = bigrams.generateSortedProbabilities(isNTP);
    	spa.exportSortedProbabilities(sbpFile);
    }
}

class BigramArrayLists {
    private ArrayList<String> words;
    private ArrayList<Double> probs;

    public BigramArrayLists() {
	words = new ArrayList<String>();
	probs = new ArrayList<Double>();
    }

    public ArrayList<String> getWords() {
	return words;
    }

    public ArrayList<Double> getProbs() {
	return probs;
    }

    public void addWordAndProb(String word, double prob) {
	words.add(word);
	probs.add(prob);
    }
}

class NamedProbability implements Comparable<NamedProbability> {
	String name;
	double value;
	
	public NamedProbability(String n, double v) {
		name = n;
		value = v;
	}
	
	String getName() {
		return name;
	}
	
	double getValue() {
		return value;
	}
	
	public int compareTo(NamedProbability other) {
		double otherVal = other.getValue();
		if(value < otherVal)
			return -1;
		else if(value > otherVal)
			return 1;
		else
			return 0;
	}
}
