import java.util.*;
import java.io.*;

class SortedProbabilityAccessor {
    private Hashtable<String, WordsAndProbabilities> outerTable;
    private WordsAndProbabilities emptyWAP;
    
    public SortedProbabilityAccessor() {
	outerTable = new Hashtable<String, WordsAndProbabilities>();
	emptyWAP = new WordsAndProbabilities(new String[0], new double[0]);
    }

    public SortedProbabilityAccessor(String filename) {
	outerTable = new Hashtable<String, WordsAndProbabilities>();

	importSortedProbabilities(filename);
    }

    public void populateOuterWord(String outerWord, String[] innerWords, double[] probs) {
	outerTable.put(outerWord, new WordsAndProbabilities(innerWords, probs));
    }

    public void importSortedProbabilities(String filename) {
	BufferedReader inputStream = null;
	try {
	    inputStream = new BufferedReader(new FileReader(filename));
	    String line;
	    // System.out.println("Loading file " + filename);
	    String outerWord = null;
	    ArrayList<String> innerWords = null;
	    ArrayList<Double> innerProbs = null;
	    while ((line = inputStream.readLine()) != null) {
		if (line.length() < 1)
		    continue;
		switch (line.charAt(0)) {
		case '\t':
		    assert(outerWord != null);
		    String[] splitLine = line.substring(1).split(" ");
		    assert(splitLine.length == 2);
		    innerWords.add(splitLine[0]);
		    double decodedProb = -1000;
		    try {
			decodedProb = Double.parseDouble(splitLine[1]);
		    } catch (NumberFormatException e) {
			System.out.println("Warning: Line contained non-numerical values: " + splitLine[1]);
		    }
		    innerProbs.add(decodedProb);
		    break;

		default:
		    // only one word on this line if it is the outer case
		    assert(line.split("\\s+").length == 1);
		    if (outerWord != null) {
			assert(innerWords.size() > 0);
			String[] innerWordsArray = (String []) innerWords.toArray(new String[innerWords.size()]);
			double[] innerProbsArray = new double[innerProbs.size()];
			for (int i = 0; i < innerProbs.size(); i++) {
			    innerProbsArray[i] = innerProbs.get(i);
			}
			populateOuterWord(outerWord, innerWordsArray, innerProbsArray);
		    }
		    innerWords = new ArrayList<String>();
		    innerProbs = new ArrayList<Double>();
		    outerWord = line;
		    break;
		}
	    }
	    // populate for the last outerWord
	    assert(outerWord != null);
	    assert(innerWords.size() > 0);
	    //String[] innerWordsArray = (String []) innerWords.toArray();
	    String[] innerWordsArray = (String []) innerWords.toArray(new String[innerWords.size()]);
	    double[] innerProbsArray = new double[innerProbs.size()];
	    for (int i = 0; i < innerProbs.size(); i++) {
		innerProbsArray[i] = innerProbs.get(i);
	    }
	    populateOuterWord(outerWord, innerWordsArray, innerProbsArray);
	    inputStream.close();
	} catch (IOException e) {
	    System.out.println("Failed to import filename: " + filename);
	}
    }

    public void exportSortedProbabilities(String filename) {
	try {
	    BufferedWriter writer = new BufferedWriter(new FileWriter(filename));
	    Enumeration en = outerTable.keys();
	    while (en.hasMoreElements()) {
		String key = (String) en.nextElement();
		WordsAndProbabilities value = outerTable.get(key);
		String keyString = key + "\n";
		writer.write(keyString, 0, keyString.length());
		String[] valueWords = value.getWords();
		double[] valueProbs = value.getProbs();
		for (int i = 0; i < valueWords.length; i++) {
		    String valueString = "\t" + valueWords[i] + " " + 
			Double.toString(valueProbs[i]) + "\n";
		    writer.write(valueString, 0, valueString.length());
		}
	    }
	    writer.close();
        } catch (IOException e) {
	    System.out.println("Failed to export to file: " + filename);
	}
    }

    public WordsAndProbabilities getInnerData(String outerWord) {
        if(!outerTable.containsKey(outerWord)) {
        	return emptyWAP;
        }
	return outerTable.get(outerWord);
    }

    public static void main(String [] args) {
        SortedProbabilityAccessor outputAccessor = new SortedProbabilityAccessor();

        String [] innerWords;
        double [] innerProbs;

        innerWords = new String[2];
        innerProbs = new double[2];
        innerWords[0] = "two";
        innerWords[1] = "words";
        innerProbs[0] = -1.45;
        innerProbs[1] = -5;

        outputAccessor.populateOuterWord("first", innerWords, innerProbs);

        innerWords = new String[2];
        innerProbs = new double[2];
        innerWords[0] = "more";
        innerWords[1] = "stuff";
        innerProbs[0] = -2.45;
        innerProbs[1] = -81.5;

        outputAccessor.populateOuterWord("second", innerWords, innerProbs);

        outputAccessor.exportSortedProbabilities("tempProbs.txt");

        SortedProbabilityAccessor inputAccessor = new SortedProbabilityAccessor();
        inputAccessor.importSortedProbabilities("tempProbs.txt");

        WordsAndProbabilities wap;
        wap = inputAccessor.getInnerData("first");
        System.out.println("first:");
        System.out.println(Arrays.toString(wap.getWords()));
        System.out.println(Arrays.toString(wap.getProbs()));

        wap = inputAccessor.getInnerData("second");
        System.out.println("second:");
        System.out.println(Arrays.toString(wap.getWords()));
        System.out.println(Arrays.toString(wap.getProbs()));

        System.out.println("Length of silly array: " + Integer.toString(inputAccessor.getInnerData("adfasdfSD").getWords().length));
    }
}

class WordsAndProbabilities {
    private String[] words;
    private double[] probs;

    public WordsAndProbabilities(String[] ws, double[] ps) {
	// could check that probs are in descending order
	words = ws;
	probs = ps;
    }

    public String getWord(int index) {
	return words[index];
    }

    public double getProb(int index) {
	return probs[index];
    }
    
    public String[] getWords() {
	return words;
    }

    public double[] getProbs() {
	return probs;
    }    
}
