
/**
 * Describes an interface for getting, setting, importing, and
 * exporting parameters used by IBM models 1, 2, and 3.
 * In this interface and its implementation(s), source and
 * dest refer to the source and destination languages of the
 * IBM generative models, not to the source and destination of
 * the overall translation system.
 * The decision for how to format parameter storage for the
 * import and export functions is implementation-specific.
 */

public abstract class ParameterAccessor {

    /**
     * Returns the T value for the specified destination and source words.
     * If a given word pair has not had a probability value stored,
     * <code>0</code> is returned.

     @param destWord the destination word for the T value
     @param sourceWord the source word for the T value

     @return the T value for this pair of words
     */

    public abstract double getT(String destWord, String sourceWord);
    
    public double getLogT(String destWord, String sourceWord) {
	double regProb = getT(destWord, sourceWord);
	if (regProb > 0.0)
	    return Math.log(regProb) / Math.log(10);
	else
	    return Double.NEGATIVE_INFINITY;			 
    }

    /**
     * Sets the T value for the specified destination and source words.

     @param destWord the destination word for the T value
     @param sourceWord the source word for the T value
     @param value the value to assign for this pair of words
     */

    public abstract void setT(String destWord, String sourceWord, double value);

    /**
     * Returns the A value for the given word position in the source
     * sentence given a position in the destination sentence and the lengths
     * of the two sentences.

     @param sourcePos the word position in the source sentence
     @param destPos the word position in the dest sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence

     @return the A value for these inputs
     */

    public abstract double getA(int sourcePos, int destPos, int sourceLength, int destLength);
    
    public double getLogA(int sourcePos, int destPos, int sourceLength, int destLength) {
	double regProb = getA(sourcePos, destPos, sourceLength, destLength);
	if (regProb > 0.0)
	    return Math.log(regProb) / Math.log(10);
	else
	    return Double.NEGATIVE_INFINITY;			 
    }
    /**
     * Sets A value for the given parameters to the given value.

     @param sourcePos the word position in the source sentence
     @param destPos the word position in the dest sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence
     @param value the value to set for this A value
     */
    
    public abstract void setA(int sourcePos, int destPos, int sourceLength, int destLength, double value);

    /**
     * Returns the D value for the given word position in the destination
     * sentence given a position in the source sentence and the lengths
     * of the two sentences.

     @param destPos the word position in the dest sentence
     @param sourcePos the word position in the source sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence

     @return the D value for these inputs
     */
    
    public abstract double getD(int destPos, int sourcePos, int sourceLength, int destLength);

    public double getLogD(int destPos, int sourcePos, int sourceLength, int destLength) {
	double regProb = getD(destPos, sourcePos, sourceLength, destLength);
	if (regProb > 0.0)
	    return Math.log(regProb) / Math.log(10);
	else
	    return Double.NEGATIVE_INFINITY;			 
    }
    /**
     * Sets the D value for the given parameters to the given value.

     @param destPos the word position in the dest sentence
     @param sourcePos the word position in the source sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence
     @param value the value to set for this D value
     */
    
    public abstract void setD(int destPos, int sourcePos, int sourceLength, int destLength, double value);

    /**
     * Returns the N value for a given fertility value and source-language word.

     @param number the fertility for the word
     @param word the source-language word
     
     @return the N value for these parameters (the probability of the
     * given word having the given fertility)
     */
    
    public abstract double getN(int number, String word);

    public double getLogN(int number, String word) {
	double regProb = getN(number, word);
	if (regProb > 0.0)
	    return Math.log(regProb) / Math.log(10);
	else
	    return Double.NEGATIVE_INFINITY;			 
    }

    /**
     * Sets the N value for a given fertility value and source-language word.

     @param number the fertility for the word
     @param word the source-language word
     @param value the N value to be set for this fertility and word
     */
    
    public abstract void setN(int number, String word, double value);

    /**
     * Returns the P value, the probability of inserting a spurious word
     * after each non-spurious word position in the destination sentence
     * being generated.

     @return the P value
     */
    
    public abstract double getP();

    public double getLogP() {
	double regProb = getP();
	if (regProb > 0.0)
	    return Math.log(regProb) / Math.log(10);
	else
	    return Double.NEGATIVE_INFINITY;			 
    }

    /**
     * Sets the P value, the probability of inserting a spurious word
     * after each non-spurious word position in the destination sentence
     * being generated.

     @param value the value to be set for the P value
     */
    
    public abstract void setP(double value);

    /**
     * Imports the T, A, D, N, and P values stored in the specified file,
     * writing them into the parameter accessor object. The parameters
     * in the file will overwrite any conflicting parameters already in
     * the object, but will not delete already-present parameter values
     * that are unspecified by the file. The format the parameter file
     * is dependent on the implementation of this interface.

     @param filename the name of the parameter file (relative path)
     */
    
    public abstract void importParams(String filename);

    /**
     * Exports the T, A, D, N, and P values to the specified file.
     * The file format used is dependent on the implementation of the
     * interface. Any file existing at the given file name will be
     * replaced.

     @param filename the name of the parameter file (relative path)
     */
    
    public abstract void exportParams(String filename);
    
}  
