
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.PriorityQueue;
import java.util.ArrayList;
import java.util.Arrays;
import java.io.FileWriter;
import java.io.FileReader;
import java.io.BufferedWriter;
import java.io.BufferedReader;
import java.io.IOException;
import java.lang.NumberFormatException;
import java.lang.Math;
import java.util.LinkedList;

/**
 * Stores a double value and an associated string, to allow the
 * strings to be sorted and printed by the ordering of the doubles.
 * Used for printing and examining the highest-probability T
 * values generated during Model 1 training.
 */

class NamedProbability implements Comparable<NamedProbability> {
    String name;
    double value;
	
    public NamedProbability(String n, double v) {
	name = n;
	value = v;
    }
	
    String getName() {
	return name;
    }
	
    double getValue() {
	return value;
    }
	
    public int compareTo(NamedProbability other) {
	double otherVal = other.getValue();
	if(value < otherVal)
	    return -1;
	else if(value > otherVal)
	    return 1;
	else
	    return 0;
    }
}

/**
 * Implements the ParameterAccessor interface to hold, provide
 * access to, import, and export parameters for IBM models 1, 2, and 3.
 * In this interface and its implementation(s), source and
 * dest refer to the source and destination languages of the
 * IBM generative models, not to the source and destination of
 * the overall translation system.
 * For importing and exporting, the parameters are stored in a
 * straightforward ASCII-encoded format, one parameter value per line.
 */

public class ParametersFromTextFile extends ParameterAccessor {
	
    Hashtable<String, Double> tableT;
    Hashtable<String, Double> tableA;
    Hashtable<String, Double> tableD;
    Hashtable<String, Double> tableN;
    Hashtable<String, Double> lastACounts;
    double p;
	
    /**
     * Initializes the data structures that hold the parameter
     * values of the object.
     */

    public ParametersFromTextFile() {
	tableT = new Hashtable<String, Double>();
	tableA = new Hashtable<String, Double>();
	tableD = new Hashtable<String, Double>();
	tableN = new Hashtable<String, Double>();
	lastACounts = new Hashtable<String, Double>();
	p = 0;
    }

    public ParametersFromTextFile(String filename) {
	tableT = new Hashtable<String, Double>();
	tableA = new Hashtable<String, Double>();
	tableD = new Hashtable<String, Double>();
	tableN = new Hashtable<String, Double>();
	lastACounts = new Hashtable<String, Double>();
	p = 0;

	importParams(filename);
    }
    
    public void setLastACounts(Hashtable<String, Double> ACounts) {
	lastACounts = ACounts;	
    }

    public Hashtable<String, Double> getLastACounts() {
	return lastACounts;
    }
	
    /**
     * Returns the T value for the specified destination and source words.
     * If a given word pair has not had a probability value stored,
     * <code>0</code> is returned.

     @param destWord the destination word for the T value
     @param sourceWord the source word for the T value

     @return the T value for this pair of words
    */

    public double getT(String destWord, String sourceWord)
    {
        if(tableT.containsKey(destWord + " " + sourceWord))
	    return tableT.get(destWord + " " + sourceWord).doubleValue();
        else
	    return 0.0;
    }

    /**
     * Sets the T value for the specified destination and source words.

     @param destWord the destination word for the T value
     @param sourceWord the source word for the T value
     @param value the value to assign for this pair of words
    */

    public void setT(String destWord, String sourceWord, double value)
    {
    	tableT.put(destWord + " " + sourceWord, new Double(value));
    }

    /**
     * Returns the A value for the given word position in the source
     * sentence given a position in the destination sentence and the lengths
     * of the two sentences.

     @param sourcePos the word position in the source sentence
     @param destPos the word position in the dest sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence

     @return the A value for these inputs
    */
    
    public double getA(int sourcePos, int destPos, int sourceLength, int destLength)
    {	
    	if(tableA.containsKey(Integer.toString(sourcePos) + " " + Integer.toString(destPos) + " " + Integer.toString(sourceLength) + " " + Integer.toString(destLength))) {
	    return tableA.get(Integer.toString(sourcePos) + " " + Integer.toString(destPos) + " " + Integer.toString(sourceLength) + " " + Integer.toString(destLength)).doubleValue();
    	} else
	    return 0.0;
    }

    /**
     * Sets A value for the given parameters to the given value.

     @param sourcePos the word position in the source sentence
     @param destPos the word position in the dest sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence
     @param value the value to set for this A value
    */
    
    public void setA(int sourcePos, int destPos, int sourceLength, int destLength, double value)
    {
    	tableA.put(Integer.toString(sourcePos) + " " + Integer.toString(destPos) + " "
		   + Integer.toString(sourceLength) + " " + Integer.toString(destLength),
		   new Double(value));
    }

    /**
     * Returns the D value for the given word position in the destination
     * sentence given a position in the source sentence and the lengths
     * of the two sentences.

     @param destPos the word position in the dest sentence
     @param sourcePos the word position in the source sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence

     @return the D value for these inputs
    */
    
    public double getD(int destPos, int sourcePos, int sourceLength, int destLength)
    {
	if (tableD.containsKey(Integer.toString(destPos) + " " + Integer.toString(sourcePos) + " "
			       + Integer.toString(sourceLength) + " " + Integer.toString(destLength))) {
	    return tableD.get(Integer.toString(destPos) + " " + Integer.toString(sourcePos) + " "
			  + Integer.toString(sourceLength) + " " + Integer.toString(destLength)).doubleValue();
	} else {
	    return 0.0;
	}
    }

    /**
     * Sets the D value for the given parameters to the given value.

     @param destPos the word position in the dest sentence
     @param sourcePos the word position in the source sentence
     @param sourceLength the length (in words) of the source sentence
     @param destLength the length (in words) of the dest sentence
     @param value the value to set for this D value
    */
    
    public void setD(int destPos, int sourcePos, int sourceLength, int destLength, double value)
    {
    	tableD.put(Integer.toString(destPos) + " " + Integer.toString(sourcePos) + " "
		   + Integer.toString(sourceLength) + " " + Integer.toString(destLength),
		   new Double(value));
    }

    /**
     * Returns the N value for a given fertility value and source-language word.

     @param number the fertility for the word
     @param word the source-language word
     
     @return the N value for these parameters (the probability of the
     * given word having the given fertility)
     */
    
    public double getN(int number, String word)
    {
    	String key = Integer.toString(number) + " " + word;
	if (tableN.containsKey(key)) { return tableN.get(key).doubleValue(); }
	else { return 0.0; }
    }

    /**
     * Sets the N value for a given fertility value and source-language word.

     @param number the fertility for the word
     @param word the source-language word
     @param value the N value to be set for this fertility and word
    */
    
    public void setN(int number, String word, double value)
    {
    	tableN.put(Integer.toString(number) + " " + word, new Double(value));
    }

    /**
     * Sets the P value, the probability of inserting a spurious word
     * after each non-spurious word position in the destination sentence
     * being generated.

     @param value the value to be set for the P value
    */
    
    public double getP()
    {
    	return p;
    }

    /**
     * Sets the P value, the probability of inserting a spurious word
     * after each non-spurious word position in the destination sentence
     * being generated.

     @param value the value to be set for the P value
    */
    
    public void setP(double value)
    {
    	p = value;
    }

    /**
     * Imports the T, A, D, N, and P values stored in the specified file,
     * writing them into the parameter accessor object. The parameters
     * in the file will overwrite any conflicting parameters already in
     * the object, but will not delete already-present parameter values
     * that are unspecified by the file. Currently, the function swallows
     * any exceptions and warns of failure only via print statements.

     @param filename the name of the parameter file (relative path)
    */
    
    public void importParams(String filename)
    {
        //ArrayList<String[]> splitLines = new ArrayList<String[]>();
        //loadToArrayList(filename, splitLines);
        
        BufferedReader inputStream = null;
	try {
	    inputStream = new BufferedReader(new FileReader(filename));
	    String line;
	    // System.out.println("Loading file " + filename);
	    while ((line = inputStream.readLine()) != null) {
		//do stuff
		String [] lineList = line.split("\\s+");
				
		if(lineList.length < 2)
		    {
			System.out.println("Warning: line with less than two elements: " + Arrays.toString(lineList));
			continue;
		    }
                if(lineList[0].length() == 0)
		    {
			System.out.println("Warning: empty first string (not sure how that's possible)");
			continue;
		    }
                switch(lineList[0].charAt(0))
		    {
                    case 'T':
                        if(lineList.length != 4)
                            System.out.println("Warning: T line had " + Integer.toString(lineList.length) + " elements: " + Arrays.toString(lineList));
                        else
			    {
				try {
				    setT(lineList[1], lineList[2], Double.parseDouble(lineList[3]));
				} catch (NumberFormatException e) {
				    System.out.println("Warning: T line had non-numerical stuff: " + Arrays.toString(lineList));
				}
			    }
                        break;
                    case 'A':
                        if(lineList.length != 6)
                            System.out.println("Warning: A line had " + Integer.toString(lineList.length) + " elements: " + Arrays.toString(lineList));
                        else
			    {
				try {
				    setA(Integer.decode(lineList[1]), Integer.decode(lineList[2]), Integer.decode(lineList[3]), Integer.decode(lineList[4]), Double.parseDouble(lineList[5]));
				} catch (NumberFormatException e) {
				    System.out.println("Warning: A line had non-numerical stuff: " + Arrays.toString(lineList));
				}
			    }
                        break;
                    case 'D':
                        if(lineList.length != 6)
                            System.out.println("Warning: D line had " + Integer.toString(lineList.length) + " elements: " + Arrays.toString(lineList));
                        else
			    {
				try {
				    setD(Integer.decode(lineList[1]), Integer.decode(lineList[2]), Integer.decode(lineList[3]), Integer.decode(lineList[4]), Double.parseDouble(lineList[5]));
				} catch (NumberFormatException e) {
				    System.out.println("Warning: D line had non-numerical stuff: " + Arrays.toString(lineList));
				}
			    }
                        break;
                    case 'N':
                        if(lineList.length != 4)
                            System.out.println("Warning: N line had " + Integer.toString(lineList.length) + " elements: " + Arrays.toString(lineList));
                        else
			    {
				try {
				    setN(Integer.decode(lineList[1]), lineList[2], Double.parseDouble(lineList[3]));
				} catch (NumberFormatException e) {
				    System.out.println("Warning: N line had non-numerical stuff: " + Arrays.toString(lineList));
				}
			    }
                        break;
                    case 'P':
                        if(lineList.length != 2)
                            System.out.println("Warning: P line had " + Integer.toString(lineList.length) + " elements: " + Arrays.toString(lineList));
                        else
			    {
				try {
				    setP(Double.parseDouble(lineList[1]));
				} catch (NumberFormatException e) {
				    System.out.println("Warning: P line had non-numerical stuff: " + Arrays.toString(lineList));
				}
			    }
                        break;
                    case 'Z':
                        if(lineList.length != 6)
                            System.out.println("Warning: Z line had " + Integer.toString(lineList.length) + " elements: " + Arrays.toString(lineList));
                        else
			    {
				try {
				    lastACounts.put((lineList[1])+" "+(lineList[2])+" "+(lineList[3])+" "+(lineList[4]), Double.parseDouble(lineList[5]));
				} catch (NumberFormatException e) {
				    System.out.println("Warning: Z line had non-numerical stuff: " + Arrays.toString(lineList));
				}
			    }
                        break;

                    default:
                        System.out.println("Warning: unparsable line: " + Arrays.toString(lineList));
                        break;
		    }
				
	    }
	    //System.out.print(arrayListToString(sentenceList));
	    inputStream.close();
	} catch (IOException e) {
	    System.out.println("Could not open " + filename);
	    return;
	}
    }

    /**
     * Exports the T, A, D, N, and P values to the specified file.
     * The file format used is dependent on the implementation of the
     * interface. Any file existing at the given file name will be
     * replaced. Currently, the function swallows any exceptions and
     * only warns of failure via print statements.

     @param filename the name of the parameter file (relative path)
    */
    
    public void exportParams(String filename)
    {
        try {
            BufferedWriter writer = new BufferedWriter(new FileWriter(filename));
            writeHashTableToFile("T", tableT, writer);
            writeHashTableToFile("A", tableA, writer);
            writeHashTableToFile("D", tableD, writer);
            writeHashTableToFile("N", tableN, writer);
	    writeHashTableToFile("Z", lastACounts, writer);
            String pString = "P " + Double.toString(p) + "\n";
            writer.write(pString, 0, pString.length());
            writer.close();
        } catch (IOException e) {
            System.out.println("Failed to write to " + filename + ": " + e.toString());
        }
    }

    /**
     * Writes the specified String-to-Double hash table to the specified
     * file object, ASCII-formatted with one entry per line.
     * The key and value are space-separated, and prepended to each
     * line is a prefix indicating the hashtable being written.
     * This performs the grunt work of exporting the parameters to
     * a file.

     @param prefix the prefix to prepend to each line
     @param table the table to be written
     @param file the file object to use for writing

     @throws IOException upon failure during writing
    */
    
    public void writeHashTableToFile(String prefix, Hashtable<String, Double> table, BufferedWriter file) throws IOException
    {
        Enumeration en = table.keys();
        while (en.hasMoreElements()) {
            String key = (String) en.nextElement();
            String value = table.get(key).toString();
            String output = prefix + " " + key + " " + value +"\n";
            file.write(output, 0, output.length()); 
        }
    }

    /**
     * Prints the stored T probabilities in ascending order, along
     * with the word pairs these values are indexed by.
     */
    
    public void printT()
    {
    	int count = 0;
    	//WARNING: remember that for now, this only prints probabilities greater than 0.0001!
    	PriorityQueue<NamedProbability> queue = new PriorityQueue<NamedProbability>();
    	Enumeration en = tableT.keys();
    	while(en.hasMoreElements())
	    {
    		String key = (String) en.nextElement();
    		queue.add(new NamedProbability(key, tableT.get(key).doubleValue()));
	    }
    	NamedProbability np = queue.poll();
    	while(np != null) {
	    if(np.getValue() > 0.0001) {
		System.out.println(np.getName() + ": " + Double.toString(np.getValue()));
		count+=1;
	    }
	    np = queue.poll();
    	}
    	System.out.println(count);
    }
    
    public void printA()
    {
    	int count = 0;
    	//WARNING: remember that for now, this only prints probabilities greater than 0.0001!
    	PriorityQueue<NamedProbability> queue = new PriorityQueue<NamedProbability>();
    	Enumeration en = tableA.keys();
    	while(en.hasMoreElements())
	    {
    		String key = (String) en.nextElement();
    		queue.add(new NamedProbability(key, tableA.get(key).doubleValue()));
	    }
    	NamedProbability np = queue.poll();
    	while(np != null) {
	    if(np.getValue() > 0.0001) {
		System.out.println(np.getName() + ": " + Double.toString(np.getValue()));
		count+=1;
	    }
	    np = queue.poll();
    	}
    	System.out.println(count);
    }

    /**
     * Performs unit tests on the parameter accessor, writing and
     * reading the file unitTestExport.txt in the process.
     */
    
    public void testSums() {
	Hashtable<String, Double> tTotal = new Hashtable<String, Double>();
	Hashtable<String, Double> dTotal = new Hashtable<String, Double>();
	Hashtable<String, Double> aTotal = new Hashtable<String, Double>();
	Hashtable<String, Double> nTotal = new Hashtable<String, Double>();
	LinkedList<Hashtable<String, Double>> tables = new LinkedList<Hashtable<String, Double>>();
	tables.add(tTotal);
	tables.add(dTotal);
	tables.add(aTotal);
	tables.add(nTotal);
	

	Enumeration en;


	en = tableT.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    String[] splitKeys = key.split(" ");
	    double value = getT(splitKeys[0], splitKeys[1]);
	    String sourceWord = splitKeys[1];
	    if (tTotal.containsKey(sourceWord)) { tTotal.put(sourceWord, tTotal.get(sourceWord)+value); }
	    else { tTotal.put(sourceWord, value); }
	}

	en = tableD.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double value = tableD.get(key);
	    String[] splitKeys = key.split(" ");
	    String totalKey = splitKeys[1]+" "+splitKeys[2]+" "+splitKeys[3];
	    if (dTotal.containsKey(totalKey)) { dTotal.put(totalKey, dTotal.get(totalKey)+value); }
	    else { dTotal.put(totalKey, value); }
	}

	en = tableA.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double value = tableA.get(key);
	    String[] splitKeys = key.split(" ");
	    String totalKey = splitKeys[1]+" "+splitKeys[2]+" "+splitKeys[3];
	    if (aTotal.containsKey(totalKey)) { aTotal.put(totalKey, aTotal.get(totalKey)+value); }
	    else { aTotal.put(totalKey, value); }
	}

	en = tableN.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    String[] splitKeys = key.split(" ");
	    double value = getN(Integer.parseInt(splitKeys[0]), splitKeys[1]);
	    String sourceWord = splitKeys[1];
	    if (nTotal.containsKey(sourceWord)) { nTotal.put(sourceWord, nTotal.get(sourceWord)+value); }
	    else { nTotal.put(sourceWord, value); }
	}

	for (Hashtable<String, Double> ht : tables) {
	    System.out.println("==================");
	    en = ht.keys();
	    while (en.hasMoreElements()) {
		String key = (String) en.nextElement();
		double value = ht.get(key);
		if (Math.abs(1-value) > .00001) {
		    System.out.print(key+" ");	       
		    System.out.println(value);
		}
	    }
	}
    }
    public static void main(String[] args)
    {
    	System.out.println("Testing get and set...");
    	
    	ParametersFromTextFile params = new ParametersFromTextFile();
    	
    	double tval = 0.42452;
    	double aval = 0.232346;
    	double dval = 0.987975;
    	double nval = 0.1234;
    	double pval = 0.6767676;

    	double logtval = Math.log(0.42452) / Math.log(10);
    	double logaval = Math.log(0.232346) / Math.log(10);
    	double logdval = Math.log(0.987975) / Math.log(10);
    	double lognval = Math.log(0.1234) / Math.log(10);
    	double logpval = Math.log(0.6767676) / Math.log(10);

    	params.setT("casa", "house", tval);
    	params.setT("casa", "car", 0.00052);
    	params.setT("iglesia", "church", 0.72452);
    	params.setT("amigo", "enemy", 0.13452);
    	params.setA(1, 4, 7, 5, aval);
    	params.setD(1, 6, 9, 13, dval);
    	params.setN(12, "hello", nval);
    	params.setP(pval);
    	
    	assert(Math.abs(params.getLogT("casa", "house") - logtval) < .0001);
    	assert(Math.abs(params.getLogA(1, 4, 7, 5) - logaval) <.0001);
    	assert(Math.abs(params.getLogD(1, 6, 9, 13) - logdval) < .0001);
    	assert(Math.abs(params.getLogN(12, "hello") - lognval) < .0001);
    	assert(Math.abs(params.getLogP() - logpval) < .0001);
    	System.out.println("Test successful");
    	
    	System.out.println("Testing export and import");
    	
    	params.exportParams("unitTestExport.txt");
    	
    	ParametersFromTextFile params2 = new ParametersFromTextFile();
    	
    	params2.importParams("unitTestExport.txt");
    	assert(params2.getT("casa", "house") == tval);
    	assert(params2.getA(1, 4, 7, 5) == aval);
    	assert(params2.getD(1, 6, 9, 13) == dval);
    	assert(params2.getN(12, "hello") == nval);
    	assert(params2.getP() == pval);
    	System.out.println("Test successful");

    }
	
}
