
import java.util.Hashtable;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;

/**
 * Implements the TranslationModel interface to perform the training
 * and sentence probability evaluation of IBM Model 2.
 */
class Model2 extends TranslationModel {
    
    //the maximum length of a sentence in the corpus; this is used in initializing A values
    //you can overestimate this, but underestimating will cause exceptions
    private final int maxLength = 15;
    
	
    public Model2()
    {}

    /**
     * Initializes the given parameters to contain uniform A values
     * for the given corpus. Assumes that the given parameter object
     * is correctly trained from Model 1.
     */
    public void initialize(ParameterAccessor params, CorpusAccessor corpus)
    {
	int count = 0;
	//initialize A values
	for (int m=1; m < maxLength; m++) 
	    {
		for (int l=1; l < maxLength; l++)
		    {
		        for (int i=1; i < l+1; i++)
			    {
				for (int j=1; j < m+1; j++)
				    {
					/*if ((j==1) && (i==6) && (m==5) && (l==6)) {
					  System.out.println("found target");
					  }*/
					params.setA(i, j, l, m, 1.0/(l));
					count++;
				    }
			    }
		    }
	    }
	//System.out.print("count: ");
	//System.out.println(count);
    }
	 
    /**
     * Performs model 2 training (called by trainAll)
     @param params set of parameters
     @param corpus for training
     @param model2Iters number of model 2 iterations
    */
    
    void train(ParameterAccessor params, CorpusAccessor corpus, int iters)
    {		    
		
	System.out.println("About to start training...");
		
	for(int i = 0; i < iters; i++)
	    {
		int numSourceWord = 0;
		System.out.println("Training Model2 translation model - Iteration " + Integer.toString(i + 1) + " of " + Integer.toString(iters));
		generateCounts((ParametersFromTextFile) params, corpus);
		    
	    }
    }


    /**
     * Adapted from pseudocode used by Jean Mark Gawro in Linguistics 581 at San Diego State University
     * http://www-rohan.sdsu.edu/~gawron/mt_plus/mt/course_core/lectures/assignment_five.pdf
     * 
     * Loops through all sentence pairs in the corpus, and examines them to create updated T/A values.
     * This implements the EM algorithm.
     @param params The set of parameters to read from and update
     @param corups The corpus to loop through
    */
    private void generateCounts(ParametersFromTextFile params, CorpusAccessor corpus)
    {
	Hashtable<String, Double> denominators = new Hashtable<String, Double>();
	double partialCount;
	Hashtable<String, Double> Dcounts = new Hashtable<String, Double>();
	Hashtable<String, Double> Tcounts = new Hashtable<String, Double>();
	Hashtable<String, Double> Ttotal = new Hashtable<String, Double>();
	Hashtable<String, Double> Dtotal = new Hashtable<String, Double>();
		
	for (int x=0; x < corpus.getSize(); x++) {
	    String[] sourceSentence = corpus.getSentence(x, true);
	    String[] destSentence = corpus.getSentence(x, false);
	    for (int j = 1; j <destSentence.length+1; j++) {
		denominators.put(destSentence[j-1], 0.0);
		for (int i = 1; i < sourceSentence.length+1; i++) {
		    double denominator = params.getT(destSentence[j-1], sourceSentence[i-1])*params.getA(i, j, sourceSentence.length, destSentence.length);
		    denominators.put(destSentence[j-1], denominators.get(destSentence[j-1])+denominator);
		    if (destSentence[j-1].equals("e")){
			//System.out.print(params.getT(destSentence[j], sourceSentence[i]));
			//System.out.print(" ");
			//System.out.println(params.getA(j, i, destSentence.length, sourceSentence.length));
		    }
		}
		for (int i = 1; i < sourceSentence.length+1; i++) {
		    double numerator = params.getT(destSentence[j-1], sourceSentence[i-1])*params.getA(i, j, sourceSentence.length, destSentence.length);
		    partialCount = numerator/denominators.get(destSentence[j-1]);
					
		    String concatWords = destSentence[j-1]+" "+sourceSentence[i-1];
		    if (!Tcounts.containsKey(concatWords)) {
			Tcounts.put(concatWords, partialCount);
		    }
		    else {
			Tcounts.put(concatWords, new Double(Tcounts.get(concatWords).doubleValue()+partialCount));
		    }
		    String concatInts = combineFourInts(i, j, sourceSentence.length, destSentence.length);
		    if (!Dcounts.containsKey(concatInts)) {
			Dcounts.put(concatInts, partialCount);
		    }
		    else {
			Dcounts.put(concatInts, new Double(Dcounts.get(concatInts).doubleValue()+partialCount));
		    }
		    if (!Ttotal.containsKey(sourceSentence[i-1])) {
			Ttotal.put(sourceSentence[i-1], partialCount);
		    }
		    else {
			Ttotal.put(sourceSentence[i-1], new Double(Ttotal.get(sourceSentence[i-1]).doubleValue()+partialCount));
		    }					
					
		    concatInts = combineThreeInts(j, sourceSentence.length, destSentence.length);
		    if (!Dtotal.containsKey(concatInts)) {
			Dtotal.put(concatInts, partialCount);
		    }
		    else {
			Dtotal.put(concatInts, new Double(Dtotal.get(concatInts).doubleValue()+partialCount));
		    }				
		}						
	    }
	}
	params.setLastACounts(Dcounts);
	//update T values
	Enumeration en = Tcounts.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double numerator = Tcounts.get(key).doubleValue();
	    String[] separated = key.split(" ");
	    String key2 = separated[1];
	    double denominator = Ttotal.get(key2).doubleValue();
	    params.setT(separated[0], separated[1], numerator/denominator);
	}
	//update A values		
	en = Dcounts.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double numerator = Dcounts.get(key).doubleValue();
	    int[] separated = separateFourInts(key);
	    String key2 = combineThreeInts(separated[1], separated[2], separated[3]);
	    double denominator = Dtotal.get(key2).doubleValue();
	    params.setA(separated[0], separated[1], separated[2], separated[3], numerator/denominator);
	}				
    }
	
    /**
     * These functions are used to combine ints into strings, which are then used 
     * as keys for hash tables
     *
     */	
    String combineFourInts(int destPos, int sourcePos, int destLen, int sourceLen) {
	String strDestPos = Integer.toString(destPos);
	String strSourcePos = Integer.toString(sourcePos);
	String strDestLen = Integer.toString(destLen);
	String strSourceLen = Integer.toString(sourceLen);
	return strDestPos+" "+strSourcePos+" "+strDestLen+" "+strSourceLen;
    }
	
    String combineThreeInts(int sourcePos, int destLen, int sourceLen) {
	String strSourcePos = Integer.toString(sourcePos);
	String strDestLen = Integer.toString(destLen);
	String strSourceLen = Integer.toString(sourceLen);
	return strSourcePos+" "+strDestLen+" "+strSourceLen;
    }	
	
    int[] separateFourInts(String fourInts) {
	String[]  temp = fourInts.split(" ");
	int[] intArray = new int[4];
	for (int i = 0; i < 4; i++) {
	    intArray[i] = Integer.parseInt(temp[i]);
	}
	return intArray;	
    }
	
    int[] separateThreeInts(String fourInts) {
	String[]  temp = fourInts.split(" ");
	int[] intArray = new int[3];
	for (int i = 0; i < 3; i++) {
	    intArray[i] = Integer.parseInt(temp[i]);
	}
	return intArray;	
    }
	
    /**
     * Returns the probability that Model 2 would generate the given
     * destination sentence as a translation of the source sentence.
     * For Model 2, this calculation is performed exactly.

     @param params the parameters used to perform the probability
     * evaluation
     @param destSentence the sentence generated by the source sentence;
     * words must be space-separated
     @param sourceSentence the sentence generating the destination
     * sentence; words must be space-separated

     @return the probability that the given source sentence
     * generates the given destination sentence under Model 2
     */
	
    public double sentenceProbability(ParameterAccessor params, String destSentence, String sourceSentence)
    {
	//Note: this probability value does not take sentence lengths into account
	String[] destWords = destSentence.split("\\s+");
	String[] sourceWords = sourceSentence.split("\\s+");
		
	return sentenceProbability(params, destWords, sourceWords);
    }

    /**
     * Returns the probability that Model 2 would generate the given
     * destination sentence as a translation of the source sentence,
     * where the two sentences are represented as arrays of individual
     * words. Being Model 2, this calculation is performed exactly.

     @param params the parameters used to perform the probability
     * evaluation
     @param destSentence the sentence generated by the source sentence,
     * as an array of words
     @param sourceSentence the sentence generating the destination
     * sentence, as an array of words

     @return the probability that the given source sentence
     * generates the given destination sentence under Model 2
     */
	
    public double sentenceProbability(ParameterAccessor params, String[] destSentence, String[] sourceSentence)
    {
	//Note: this probability value does not take sentence lengths into account
	
	double product = 1.0;
	String destWord;
	String sourceWord;
		
	for(int j=0; j < destSentence.length; j++)
	    {
		destWord = destSentence[j];
		double sum = 0.0;
		for(int i=0; i < sourceSentence.length; i++)
		    {
		    	sourceWord = sourceSentence[i];
		        sum += params.getT(destWord, sourceWord)*params.getA(j,i,destSentence.length,sourceSentence.length);
		    }
		product *= sum;
	    }
		
	return product;
    }
	
    public static void main(String[] args)
    {
	//test initialize
		
	ParametersFromTextFile params = new ParametersFromTextFile();
	Model1 model1 = new Model1();
	Model2 model2 = new Model2();
		
	CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.es", "../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.en");
	//CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/thousand.es", "../corpora/textpairs/thousand.en");
	//CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/manual_test_corpus.es", "../corpora/textpairs/manual_test_corpus.en");

	model1.initialize(params, corpus);
        
	model1.train(params, corpus, 5);
        
	model2.initialize(params, corpus);

	model2.train(params, corpus, 5);

	//params.printT();
	//params.printA();
	params.exportParams("../parameters/textfiles/europarl_2-to-10_stripped.training.es-en.model2.5iters");

    }	
}