import java.util.*;
import java.lang.Math;

/**
 * Implements the TranslationModel interface to perform the training
 * and sentence probability evaluation of IBM Model 3.
 */
class Model3 extends TranslationModel {
    
    //the maximum length of a sentence in the corpus
    //you can overestimate this, but underestimating will cause exceptions
    private final int maxLength = 15;
    private final int maxFert = 15;    //the maximum fertility--should likely be set to the maximum length
    private final double initialP1 = .3;     //the initial P value--the probability of adding a spurious word
    private String[] UniqueWordsListSource;  //all words in the source language
    private String[] UniqueWordsListDest;    //all words in the dest language
    private final double smoothingValue = 1E-30;     //optional value to smooth all parameter values   
	
    public Model3()
    {}  

    /**
     * Converts A values from Model2 into D values for Model3
     @param parameters with A values
    */
    private void setInitialDValues(ParametersFromTextFile params) {
	Hashtable<String, Double> initialDCounts = params.getLastACounts();
	Hashtable<String, Double> initialDTotal = new Hashtable<String, Double>();
	Enumeration en = initialDCounts.keys();
	while (en.hasMoreElements()) {
	    String AKey = (String) en.nextElement();
	    double AValue = initialDCounts.get(AKey);
	    int[] AKeySep = separateFourInts(AKey);
	    String totalKey = combineThreeInts(AKeySep[0],AKeySep[2],AKeySep[3]);
	    if (initialDTotal.containsKey(totalKey)) { initialDTotal.put(totalKey, initialDTotal.get(totalKey) + AValue); }
	    else { initialDTotal.put(totalKey, AValue); }	    
	}

	en = initialDCounts.keys();
	while (en.hasMoreElements()) {
	    String AKey = (String) en.nextElement();
	    double AValue = initialDCounts.get(AKey);
	    int[] AKeySep = separateFourInts(AKey);
	    String totalKey = combineThreeInts(AKeySep[0],AKeySep[2],AKeySep[3]);
	    params.setD(AKeySep[1], AKeySep[0], AKeySep[2], AKeySep[3], AValue/initialDTotal.get(totalKey));
	}
    }

    /**
     * Given a sentence pair, finds a large number of likely alignments
     * See Brown et al. equation 44
     @param source sentence
     @param dest sentence
     @param params, should have T,D,N,P
     @return a hashset of alignments
    */
    private HashSet<Alignment> candidateAlignments(String[] sourceSentence, String[] destSentence, ParameterAccessor params) {
	Alignment model2Viterbi = Viterbi.viterbi(sourceSentence, destSentence, params);
	Alignment model3Viterbi = Viterbi.model3ViterbiApprox(model2Viterbi, sourceSentence, destSentence, (Model3Params) params);
	HashSet<Alignment> bestCandidates = new HashSet<Alignment>(5000);
	bestCandidates.add(model3Viterbi);
	Alignment[] candidates = Viterbi.findNeighbors(model3Viterbi, sourceSentence.length);

	//pegged alignments are found
	for (int i=1; i < sourceSentence.length+1; i++) {
	    for (int j=0; j < destSentence.length; j++) {
		Alignment peggedModel2Viterbi = Viterbi.viterbi(sourceSentence, destSentence, params, i, j);
		if (peggedModel2Viterbi.probability == 0){
		    continue;
		}
		Alignment peggedModel3Viterbi = Viterbi.model3ViterbiApprox(peggedModel2Viterbi, sourceSentence, destSentence, (Model3Params) params, i, j);
		Alignment[] peggedSet = Viterbi.findNeighbors(peggedModel3Viterbi, sourceSentence.length);
		bestCandidates.add(peggedModel3Viterbi);
		candidates = Viterbi.concat(peggedSet, candidates);
	    }
	}
	
	for (int i=0; i < candidates.length; i++) {
	    bestCandidates.add(candidates[i]);
	}
	return bestCandidates;
    }

    /**
     * Given a sentence pair and a bunch of alignments for that pair, finds the sum of all alignment scores P(a,f|e)
     @param source sentence
     @param dest sentence
     @param number of alignments between source and dest
     @param params, should have T,D,N,P
     @return the sum of all alignment scores
    */
    private double totalAlignmentProbability(String[] sourceSentence, String[] destSentence, HashSet<Alignment> neighborhood, ParameterAccessor params) {
	double totalProbability = 0.0;
	Iterator it = neighborhood.iterator();
	while (it.hasNext()) {
	    Alignment neighbor = (Alignment) it.next();
	    totalProbability += Viterbi.getAlignmentAndTranslationProb(neighbor, destSentence, sourceSentence, (Model3Params) params);
	}
	return totalProbability;
    }
 
    /**
     * Initializes N values to be uniform
     * that is, N(x|e) sums to 1 for all e and N(x|e) = N(y|e)
     @param params
     @param corpus
    */
    public void uniformNValues(ParameterAccessor params, CorpusAccessor corpus) {
	Hashtable<String,Integer> nCounts = new Hashtable<String, Integer>();
	Hashtable<String,Integer> nTotal = new Hashtable<String, Integer>();
	String[] sourceWords = corpus.getWordList(true);
	for (int i=0; i < sourceWords.length; i++) {
	    String sourceWord = sourceWords[i];
	    nTotal.put(sourceWord, maxFert);
	    for (int f=0; f < maxFert; f++) {
		String fertility = Integer.toString(f);
		nCounts.put(fertility+" "+sourceWord, 1);
	    }
	}

	Enumeration en = nCounts.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    String sourceWord = key.split(" ")[1];
	    int fertility = Integer.parseInt(key.split(" ")[0]);
	    double value = nCounts.get(key)/( (double) nTotal.get(sourceWord));
	    params.setN(fertility, sourceWord, value);
	}
    }

    /**
     * As Model2 does not concern itself with the null word, this function initializes T values 
     * for the null word to be used in Model3
     @param params
     @param corpus
    */
    private void initializeNullTValues(ParameterAccessor params, CorpusAccessor corpus) {
	Hashtable<String, Integer> wordCounts = new Hashtable<String, Integer>();
	double total = 0.0;
	for (int x=0; x < corpus.getSize(); x++) {
	    String[] destSentence = corpus.getSentence(x, false);
	    for (int j=0; j < destSentence.length; j++) {
		if (wordCounts.containsKey(destSentence[j])) {
		    wordCounts.put(destSentence[j], wordCounts.get(destSentence[j])+1);
		}
		else {
		    wordCounts.put(destSentence[j], 1);
		}
		total++;
	    }
	}
	Enumeration en = wordCounts.keys();
	//double sum = 0.0;
	while (en.hasMoreElements()) {
	    String word = (String) en.nextElement();
	    params.setT(word, "_N_U_L_L_", wordCounts.get(word)/total);
	    //sum += wordCounts.get(word)/total;
	}
    }
    
    /**
     * Calls various initialization functions: setInitialDValues(), uniformNValues(), and initializeNullTValues()
     * also sets the P value
     @param params
     @param corpus
    */
    public void initialize(ParameterAccessor params, CorpusAccessor corpus)
    {
	setInitialDValues((ParametersFromTextFile) params);
	params.setP(initialP1);
	uniformNValues(params, corpus);
	initializeNullTValues(params, corpus);
    }

	//The below code is a buggy implementation of pseudocode from Knight for initializing N values
	//Can be found in section 35 of Kevin Knight's workbook
	/*
	  System.out.println(corpus.getWordList(true).length);
	  //Set initial fertility values using pseudo-code from Knight
	  Hashtable<String,Double> nCounts = new Hashtable<String, Double>();
	  Hashtable<String,Double> nTotal = new Hashtable<String, Double>();
	  Hashtable<String,Double> alpha = new Hashtable<String, Double>();
	  Hashtable<String,Double> sums = new Hashtable<String, Double>();
	  for (int x=0; x < corpus.getSize(); x++) {
	  String[] sourceSentence = corpus.getSentence(x, true);
	  String[] destSentence = corpus.getSentence(x, false);
			
	  for (int i=1; i < sourceSentence.length+1; i++) {
				
	  for (int k=1; k < maxFert; k++) {
	  double beta = 0.0;
					
	  for (int j=1; j < destSentence.length; j++) {
	  beta += Math.pow(params.getT(destSentence[j-1],sourceSentence[i-1])*params.getA(j, i, destSentence.length, sourceSentence.length)
	  /(1 - params.getT(destSentence[j-1],sourceSentence[i-1])*params.getA(j, i, destSentence.length, sourceSentence.length)), k);
	  }
	  alpha.put(Integer.toString(i)+" "+Integer.toString(k), Math.pow(-1, k+1)*beta/k);
	  }
	  }
			
	  for (int i=1; i < sourceSentence.length+1; i++) {
	  double r = 1.0;
				
	  for (int j=1; j < destSentence.length+1; j++) {
	  r*=(1-params.getT(destSentence[j-1],sourceSentence[i-1])*params.getA(j, i, destSentence.length, sourceSentence.length));
	  }
	  for (int phi = 0; phi < maxFert; phi++) {
	  double sum = 0.0;
					
	  ArrayList<ArrayList<Integer>> bigGamma = Viterbi.generatePartitions(phi);
	  int len = bigGamma.size();
	  for (int g=0; g < len; g++) {
	  double prod = 1.0;
						
	  for (int k=1; k < phi; k++) {
	  prod*=Math.pow(alpha.get(Integer.toString(i)+" "+Integer.toString(k)),Viterbi.arrayListCount(bigGamma.get(g),k))/Viterbi.factorial(Viterbi.arrayListCount(bigGamma.get(g),k));
	  }
	  sum += prod;
	  }
	  if (nCounts.containsKey(Integer.toString(phi)+" "+sourceSentence[i-1])) {
	  nCounts.put(Integer.toString(phi)+" "+sourceSentence[i-1], nCounts.get(Integer.toString(phi)+" "+sourceSentence[i-1])+r*sum);
	  }
	  else {
	  nCounts.put(Integer.toString(phi)+" "+sourceSentence[i-1], r*sum);
	  }
	  if (nTotal.containsKey(sourceSentence[i-1])) {
	  nTotal.put(sourceSentence[i-1], nTotal.get(sourceSentence[i-1])+r*sum);
	  }
	  else {
	  nTotal.put(sourceSentence[i-1], r*sum);
	  }
	  }
	  }
	  }
	  System.out.println(nCounts.size());
	  Enumeration en = nCounts.keys();
	  while (en.hasMoreElements()) {
	  String nCountKey = (String) en.nextElement();
	  String[] keyArray = nCountKey.split(" ");
	  int fertility = Integer.parseInt(keyArray[0]);
	  String sourceWord = keyArray[1];
	  double nCount = nCounts.get(nCountKey);
	  double total = nTotal.get(sourceWord);
	  if (sums.containsKey(sourceWord)) {
	  sums.put(sourceWord, sums.get(sourceWord)+(nCount/total));
	  }
	  else {
	  sums.put(sourceWord, nCount/total);
	  }
	  //if (nCount/total!=1) {
	  //    System.out.println(nCount/total);
	  //}
	  params.setN(fertility, sourceWord, nCount/total);
	  }
	  /*
	  en = sums.keys();
	  while (en.hasMoreElements()) {
	  System.out.print(sums.get(en.nextElement()));
	  System.out.print(" ");
	  }*/
	 
    /**
     * Performs model 3 training
     @param params set of parameters
     @param corpus for training
     @param number of iterations
    */    
    void train(ParameterAccessor params, CorpusAccessor corpus, int iters)
    {
		
	System.out.println("About to start training...");
	UniqueWordsListSource = corpus.getWordList(true);
	UniqueWordsListDest = corpus.getWordList(false);
		
	for(int i = 0; i < iters; i++)
	    {
		int numSourceWord = 0;
		System.out.println("Training Model3 translation model - Iteration " + Integer.toString(i + 1) + " of " + Integer.toString(iters));
		generateCounts((Model3Params) params, corpus);
		//params.exportParams("model3-iter-patched-smoothing-french-"+Integer.toString(i));		    
	    }
    }

    /**
     * 
     * Loops through all sentence pairs in the corpus, and examines them to create updated T/A/D/N/P values.
     * This implements the EM algorithm.
     @param params The set of parameters to read from and update
     @param corups The corpus to loop through
    */
    private void generateCounts(Model3Params params, CorpusAccessor corpus)
    {

	//the "counts" hashtables store partial counts for each parameter value
	//the "total" hashtables store the sums of these counts, and are used in normalization
	Hashtable<String, Double> Dcounts = new Hashtable<String, Double>();
	Hashtable<String, Double> Tcounts = new Hashtable<String, Double>();
	Hashtable<String, Double> Ncounts = new Hashtable<String, Double>();
	Hashtable<String, Double> Acounts = new Hashtable<String, Double>();

	Hashtable<String, Double> Ttotal = new Hashtable<String, Double>();
	Hashtable<String, Double> Dtotal = new Hashtable<String, Double>();
	Hashtable<String, Double> Ntotal = new Hashtable<String, Double>();
	Hashtable<String, Double> Atotal = new Hashtable<String, Double>();

	double p0Count = 0.0;
	double p1Count = 0.0;
	
	//implements smoothing
	for (String sourceWord: UniqueWordsListSource) {
	    Ttotal.put(sourceWord, smoothingValue*UniqueWordsListDest.length);
	    Ntotal.put(sourceWord, smoothingValue*maxFert);
	}
	Ttotal.put("_N_U_L_L_", smoothingValue*UniqueWordsListDest.length);
	
	//loops through the corpus
	for (int x=0; x < corpus.getSize(); x++) {
	    if (x%100 == 0) {
		System.out.println(x);
	    }
	    String[] sourceSentence = corpus.getSentence(x, true);
	    String[] destSentence = corpus.getSentence(x, false);
	    
	    HashSet<Alignment> candidates = candidateAlignments(sourceSentence, destSentence, params);
	    double denominator = 0.0;
	    Hashtable<Alignment, Double> alignmentProbs = new Hashtable<Alignment, Double>();
	    Iterator it = candidates.iterator();
	    while (it.hasNext()) {
		Alignment a = (Alignment) it.next();
		double prob = Viterbi.getAlignmentAndTranslationProb(a, destSentence, sourceSentence, params);
		denominator += prob;
		alignmentProbs.put(a, prob);
	    }
	    it = candidates.iterator();

	    //loops through the alignments for a given sentence pair
	    //the score of this alignment is used in updating paramter values
	    while (it.hasNext()) {
		Alignment a = (Alignment) it.next();
		double alignmentProb = alignmentProbs.get(a)/denominator;

		Hashtable<String, Double> nestedTcounts = new Hashtable<String, Double>();
		
		for (int j=0; j < destSentence.length; j++) {
		    
		    for (int i=0; i < sourceSentence.length+1; i++) {
			String sourceWord;
			if (i == sourceSentence.length) { sourceWord = "_N_U_L_L_"; }
			else { sourceWord = sourceSentence[i]; }
			//update T,D,A inside this 'if'
			if (sourceWord.equals(a.getSourceWord(j, sourceSentence))) {
			    //update T
			    String concatWords = destSentence[j]+" "+sourceWord;
			    if (Tcounts.containsKey(concatWords)) { Tcounts.put(concatWords, Tcounts.get(concatWords)+alignmentProb); }
			    else { 
				Tcounts.put(concatWords, alignmentProb);
				Ttotal.put(sourceWord, Ttotal.get(sourceWord)-smoothingValue);
			    }
			    Ttotal.put(sourceWord, Ttotal.get(sourceWord)+alignmentProb);
			    
			    int l = sourceSentence.length;
			    int m = destSentence.length;			    
			    //update D
			    concatWords = combineFourInts(j+1, i+1, l, m);
			    String totalConcatWords = combineThreeInts(i+1, l, m);
			    if (Dcounts.containsKey(concatWords)) { Dcounts.put(concatWords, Dcounts.get(concatWords)+alignmentProb); }
			    else { Dcounts.put(concatWords, alignmentProb); }
			    if (Dtotal.containsKey(totalConcatWords)) { Dtotal.put(totalConcatWords, Dtotal.get(totalConcatWords)+alignmentProb); }
			    else { Dtotal.put(totalConcatWords, alignmentProb); }
			    
			    //update A
			    concatWords = combineFourInts(i+1, j+1, l, m);
			    totalConcatWords = combineThreeInts(j+1, l, m);
			    if (Acounts.containsKey(concatWords)) { Acounts.put(concatWords, Acounts.get(concatWords)+alignmentProb); }
			    else { Acounts.put(concatWords, alignmentProb); }
			    if (Atotal.containsKey(totalConcatWords)) { Atotal.put(totalConcatWords, Atotal.get(totalConcatWords)+alignmentProb); }
			    else { Atotal.put(totalConcatWords, alignmentProb); }
			}
		    }
		}
		//update N
		for (int i=0; i < sourceSentence.length; i++) {
		    int phi = a.getFertility(i+1);
		    String sourceWord = sourceSentence[i];
		    String key = Integer.toString(phi)+" "+sourceWord;
		    if (Ncounts.containsKey(key)) { Ncounts.put(key, Ncounts.get(key)+alignmentProb); }
		    else { 
			Ncounts.put(key, alignmentProb);
			Ntotal.put(sourceWord, Ntotal.get(sourceWord)-smoothingValue);
		    }
		    Ntotal.put(sourceWord, Ntotal.get(sourceWord)+alignmentProb);
		}
		//update P
		int nullFertility = a.getFertility(0);
		p0Count += alignmentProb*(destSentence.length-2*nullFertility);
		p1Count += alignmentProb*nullFertility;
	    }	    
	}
	//mostly for debugging--we're sure to overwrite all tables
	params.tableN.clear();
	params.tableA.clear();
	params.tableD.clear();
	params.setP(0.0);
	
	//The next bunch of enumeration loops normalize a parameter value and store it
        Enumeration en = Tcounts.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double tCount = Tcounts.get(key);
	    String[] splitKey = key.split(" ");
	    double denominator = Ttotal.get(splitKey[1]);
	    params.setT(splitKey[0], splitKey[1], tCount/denominator);
	}
	en = Dcounts.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double dCount = Dcounts.get(key);
	    String[] splitKey = key.split(" ");
	    double denominator = Dtotal.get(splitKey[1]+" "+splitKey[2]+" "+splitKey[3]);
	    int[] separateInts = separateFourInts(key);
	    params.setD(separateInts[0], separateInts[1],separateInts[2],separateInts[3],dCount/denominator);
	}
	en = Acounts.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double aCount = Acounts.get(key);
	    String[] splitKey = key.split(" ");
	    double denominator = Atotal.get(splitKey[1]+" "+splitKey[2]+" "+splitKey[3]);
	    int[] separateInts = separateFourInts(key);
	    params.setA(separateInts[0], separateInts[1],separateInts[2],separateInts[3],aCount/denominator);
	}
	en = Ncounts.keys();
	while (en.hasMoreElements()) {
	    String key = (String) en.nextElement();
	    double nCount = Ncounts.get(key);
	    String[] splitKey = key.split(" ");
	    double denominator = Ntotal.get(splitKey[1]);
	    params.setN(Integer.parseInt(splitKey[0]), splitKey[1], nCount/denominator);
	}
	double pTotal = p1Count+p0Count;
	double p1 = p1Count/pTotal;
	double p0 = p0Count/pTotal;
	params.setP(p1);

	params.Ttotal = Ttotal;
	params.Ntotal = Ntotal;
	params.testSums();
    }
	
    /**
     * These functions are used to combine ints into strings, which are then used 
     * as keys for hash tables
     *
     */	
    String combineFourInts(int destPos, int sourcePos, int destLen, int sourceLen) {
	String strDestPos = Integer.toString(destPos);
	String strSourcePos = Integer.toString(sourcePos);
	String strDestLen = Integer.toString(destLen);
	String strSourceLen = Integer.toString(sourceLen);
	return strDestPos+" "+strSourcePos+" "+strDestLen+" "+strSourceLen	;
    }
	
    String combineThreeInts(int sourcePos, int destLen, int sourceLen) {
	String strSourcePos = Integer.toString(sourcePos);
	String strDestLen = Integer.toString(destLen);
	String strSourceLen = Integer.toString(sourceLen);
	return strSourcePos+" "+strDestLen+" "+strSourceLen;
    }	
	
    int[] separateFourInts(String fourInts) {
	String[]  temp = fourInts.split(" ");
	int[] intArray = new int[4];
	for (int i = 0; i < 4; i++) {
	    intArray[i] = Integer.parseInt(temp[i]);
	}
	return intArray;	
    }
	
    int[] separateThreeInts(String fourInts) {
	String[]  temp = fourInts.split(" ");
	int[] intArray = new int[3];
	for (int i = 0; i < 3; i++) {
	    intArray[i] = Integer.parseInt(temp[i]);
	}
	return intArray;	
    }
	

    //these actually don't matter at all, so they're blank
    public double sentenceProbability(ParameterAccessor params, String destSentence, String sourceSentence)
    {return 0.0;}  
    public double sentenceProbability(ParameterAccessor params, String[] destSentence, String[] sourceSentence)
    {return 0.0;}
	
    public static void main(String[] args) {
	//test initialize
	Model3Params params = new Model3Params();
	Model1 model1 = new Model1();
	Model2 model2 = new Model2();
	Model3 model3 = new Model3();
		
	//CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/europarl_2-to-10_stripped.training.fr-en.fr", "../corpora/textpairs/europarl_2-to-10_stripped.training.fr-en.en");
	CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/thousand.es", "../corpora/textpairs/thousand.en");
	//CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/manual_test_corpus.es", "../corpora/textpairs/manual_test_corpus.en");
	//CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/hundred.es", "../corpora/textpairs/hundred.en");

        model1.initialize(params, corpus);        
	model1.train(params, corpus, 5);
        
        model2.initialize(params, corpus);
        model2.train(params, corpus, 5);
	//params.exportParams("model2latest");
        //System.out.println("Testing params before training");
	//	params.testSums();
	//params.importParams("model2tesxt");
	model3.initialize(params, corpus);
	//params.importParams("model3-iter-thous-test-0");
	//System.out.println("Testing params before training");
	//params.testSums();
	model3.train(params, corpus, 5);
	//params.exportParams("model3latest");
	//System.out.println("Testing params after training");
	//params.testSums();

    }	
}