
// NOTE: in this class and all implementations, unless noted otherwise all references to "source" and "dest"
// are with respect to the GENERATIVE MODEL!!!

/**
 * Describes a common interface for the IBM models 1, 2, and 3.
 * Implementations do not maintain their own set of parameters,
 * instead performing each operation on the given ParameterAccessor
 * and CorpusAccessor instances; this allows callers to perform
 * the various operations involved in training and probability
 * evaluation piecemeal, so they may be done one at a time if
 * they are slow, importing and exporting between them.
 * This abstract class and all implementations use source
 * and dest to refer to the source and destination languages
 * of the generative model.
 */

abstract class TranslationModel {

	/**
	 * Performs initialization on the given parameters using the
	 * given corpus, making them ready for training.
	 * Depending on the implementation, parameters to be initialized
	 * by this function may need to have been previously initialized
	 * and trained by other implementations of this class
	 * (for example, lower-numbered IBM models).
	 * After this function, the given parameters/corpus pair may be
	 * trained for however many iterations.

	 @param params the set of parameters to initialize. Depending on
	 * the model, the parameter object may be assumed to be empty,
	 * or its existing contents (generated by the previous IBM model
	 * in the chain) may be needed to generate initial parameters for
	 * this model.
	 @param corpus the corpus to use in initializing the parameters
	 */

	abstract void initialize(ParameterAccessor params, CorpusAccessor corpus);

	/**
	 * Performs a chosen number of iterations of training on the
	 * given parameters, based upon the contents of a corpus.
	 * The provided parameters are assumed to have been initialized
	 * appropriately 
	 * For all implementations, performing a total of <code>n</code>
	 * iterations of training on the same parameters and corpus
	 * across several calls produces identical results to making
	 * a single call that performs <code>n</code> iterations.

	 @param params the parameters which are trained and updated
	 @param corpus the corpus used to guide the training
	 @param iters the number of iterations of training to perform
	 */
	
	abstract void train(ParameterAccessor params, CorpusAccessor corpus, int iters);

	/**
	 * Evaluates the probability that a sentence in the source
	 * language generates a sentence in the destination language,
	 * given the provided parameter values. Implementations are not
	 * guaranteed to perform this calculation exactly, since this
	 * cannot efficiently be done past Model 1.

	 @param params the parameters used to perform the probability
	 * evaluation
	 @param destSentence the sentence generated by the source sentence;
	 * words must be space-separated
	 @param sourceSentence the sentence generating the destination
	 * sentence; words must be space-separated

	 @return the probability that the given source sentence
	 * generates the given destination sentence
	 */
	
	abstract double sentenceProbability(ParameterAccessor params, String destSentence, String sourceSentence);
	
	//abstract double sentenceAndAlignmentProbability(ParameterAccessor params, String s1, Alignment a, String s2);
	
}
