import java.util.Hashtable;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.ArrayList;
import java.lang.Math;
import org.apache.commons.math.util.MathUtils;

/**
 * A utilities class used in Model3 and ASDecoder
 * Contains tools for managing alignments and finding Viterbi (optimal)
 * alignments.
 */
class Viterbi {
    public Viterbi() {}

    /**
     * Concatenates two arrays, returns the result
     * Taken from: http://stackoverflow.com/a/784842
     @param first array to concatenate
     @param second array to concatenate
     @return the result of concatenation
    */
    public static <T> T[] concat(T[] first, T[] second) {
	T[] result = Arrays.copyOf(first, first.length + second.length);
	System.arraycopy(second, 0, result, first.length, second.length);
	return result;
    }
    
    /**
     * Counts the number of times an integer appears in an arrayList
     * Not currently used
     @param the arrayList to search
     @param the integer to count
     @return the count
    */
    public static int arrayListCount(ArrayList<Integer> arrayList, int query) {
    	int count = 0;
    	int len = arrayList.size();
    	for (int i = 0; i < len; i++) {
	    if (query==arrayList.get(i)) {
		count++;
	    }
    	}
    	return count;
    }

    /**
     * Generates the ways to partition an integer
     @param the integer to partition
     @return arraylist of partitions (a partition is an arraylist)
    */
    public static  ArrayList<ArrayList<Integer>> generatePartitions(int n) {
    	if (n==0) {
	    return new ArrayList<ArrayList<Integer>>();
    	}
    	if (n == 1) {
	    ArrayList<ArrayList<Integer>> newPartitions = new ArrayList<ArrayList<Integer>>();
	    ArrayList<Integer> newPartition = new ArrayList<Integer>();
	    newPartition.add(1);
	    newPartitions.add(newPartition);
	    return newPartitions;
    	}
    	
    	ArrayList<ArrayList<Integer>> partitions = new ArrayList<ArrayList<Integer>>();
    	ArrayList<ArrayList<Integer>> oldPartitions = generatePartitions(n-1);
    	int len = oldPartitions.size();
    	for (int i = 0; i < len; i++) {
	    ArrayList<Integer> temp = (ArrayList<Integer>) oldPartitions.get(i).clone();
	    temp.add(0,1);
	    partitions.add(temp);
	    temp = (ArrayList<Integer>) oldPartitions.get(i).clone();
	    if ((temp.size() > 0) && ((temp.size() < 2) || (temp.get(1) > temp.get(0)))) {
		ArrayList<Integer> secondPart = new ArrayList(temp.subList(1,temp.size()));
		Integer firstElement = temp.get(0)+1;
		secondPart.add(0,firstElement);
		partitions.add(secondPart);
	    }
    	}
    	return partitions;
    }

    /**
     * Given an alignment and a neighbor generated from that alignment, as well as the position
     * at which they differ, returns the probability of the neighbor alignment. Note that the 
     * original alignment must have its probability attribute set. 
     * Brown et al; equation 43
     @param neighbor alignment
     @param original alignment
     @param position at which they differ
     @param source sentence that the alignments refer to
     @param dest sentence that the alignments refer to
     @param parameters, with T,D,N,P 
     @return the probability of the neighbor alignment
    */	
    public static double neighborProbability(Alignment neighborAlignment, Alignment originalAlignment, int changedDestPosition, String[] sourceSentence, String[] destSentence, Model3Params params) {

	int originalSourcePos = originalAlignment.getAlignment()[changedDestPosition];
	int neighborSourcePos = neighborAlignment.getAlignment()[changedDestPosition];

	int originalFertFromOrigAlignment = originalAlignment.getFertility(originalSourcePos);
	int neighborFertFromNeighborAlignment = neighborAlignment.getFertility(neighborSourcePos);
	int neighborFertFromOrigAlignment = originalAlignment.getFertility(neighborSourcePos);
	int originalFertFromNeighborAlignment = neighborAlignment.getFertility(originalSourcePos);

	String originalSourceWord = originalAlignment.getSourceWord(changedDestPosition, sourceSentence);
	String neighborSourceWord = neighborAlignment.getSourceWord(changedDestPosition, sourceSentence);

	String destWord = destSentence[changedDestPosition];

	double originalProb = originalAlignment.probability;

	double factor3 = ((params.getModel3T(destWord, neighborSourceWord))/(params.getModel3T(destWord, originalSourceWord)));
	double factor0;
	double factor1;
	double factor2;
	double factor4;
	double p1 = params.getP();
	double p0 = 1-p1;
	int m = destSentence.length;

	//going from a null to non-null
	if (originalSourceWord.equals("_N_U_L_L_")) {
	    factor0 = (neighborFertFromNeighborAlignment);
	    factor1 = ((params.getModel3N(neighborFertFromNeighborAlignment,neighborSourceWord))/(params.getModel3N(neighborFertFromOrigAlignment,neighborSourceWord)));
	    factor2 = (MathUtils.binomialCoefficient(m-originalFertFromNeighborAlignment, originalFertFromNeighborAlignment)*Math.pow(p0, m-2*originalFertFromNeighborAlignment)*Math.pow(p1, originalFertFromNeighborAlignment))/(MathUtils.binomialCoefficient(m-originalFertFromOrigAlignment, originalFertFromOrigAlignment)*Math.pow(p0, m-2*originalFertFromOrigAlignment)*Math.pow(p1, originalFertFromOrigAlignment));
	    factor4 = (params.getD(changedDestPosition+1, neighborSourcePos, sourceSentence.length, destSentence.length));///(1.0/m));
	}
	//non-null to null
	else if (neighborSourceWord.equals("_N_U_L_L_")) {
	    factor0 = (1.0/originalFertFromOrigAlignment);
	    factor1 = (MathUtils.binomialCoefficient(m-neighborFertFromNeighborAlignment, neighborFertFromNeighborAlignment)*Math.pow(p0, m-2*neighborFertFromNeighborAlignment)*Math.pow(p1, neighborFertFromNeighborAlignment))/(MathUtils.binomialCoefficient(m-neighborFertFromOrigAlignment, neighborFertFromOrigAlignment)*Math.pow(p0, m-2*neighborFertFromOrigAlignment)*Math.pow(p1, neighborFertFromOrigAlignment));
	    factor2 = ((params.getModel3N(originalFertFromNeighborAlignment,originalSourceWord))/(params.getModel3N(originalFertFromOrigAlignment,originalSourceWord)));
	    factor4 = (1.0)/(params.getD(changedDestPosition+1, originalSourcePos, sourceSentence.length, destSentence.length));
	}
	//no nulls involved at all
	else {
	    factor0 = (neighborFertFromNeighborAlignment/originalFertFromOrigAlignment);
	    factor1 = ((params.getModel3N(neighborFertFromNeighborAlignment,neighborSourceWord))/(params.getModel3N(neighborFertFromOrigAlignment,neighborSourceWord)));
	    factor2 = ((params.getModel3N(originalFertFromNeighborAlignment,originalSourceWord))/(params.getModel3N(originalFertFromOrigAlignment,originalSourceWord)));
	    factor4 = ((params.getD(changedDestPosition+1, neighborSourcePos, sourceSentence.length, destSentence.length))/
		       (params.getD(changedDestPosition+1, originalSourcePos, sourceSentence.length, destSentence.length)));
	}    				
	//assert !Double.isNaN(factor0);
	//assert !Double.isNaN(factor1);
	//assert !Double.isNaN(factor2);
	//assert !Double.isNaN(factor3);
	//assert !Double.isNaN(factor4);
	//assert !Double.isNaN(originalAlignment.probability);

	double adjustmentFactor = factor0*factor1*factor2*factor3*factor4;
	//if the result is illegal, return 0.0
	if ((originalProb * adjustmentFactor > 1.0) || (Double.isNaN(originalProb*adjustmentFactor))) {
	    return 0.0;
	}
	assert originalProb * adjustmentFactor <= 1.0;
	return originalProb * adjustmentFactor;
	
    }
 
    /**
     * Finds the best alignment between two sentences according to model 2, given a pegged position
     * Brown et al. equation 84
     @param the source sentence
     @param the dest sentence
     @param parameters, should have T,A
     @param pegged source position
     @param pegged dest position
     @return the best alignment
    */
    public static Alignment viterbi(String[] sourceSentence, String[] destSentence, ParameterAccessor params, int peggedSource, int peggedDest) {
	int[] alignmentArray = new int[destSentence.length];
	Alignment a = new Alignment();
	a.probability = params.getT(destSentence[peggedDest], sourceSentence[peggedSource-1])*params.getA(peggedSource, peggedDest+1, sourceSentence.length, destSentence.length);
	alignmentArray[peggedDest] = peggedSource;
	assert peggedSource != 0;
	for (int j = 0; j < destSentence.length; j++) {
	    
	    if (j != peggedDest) {
		int best_i = -1;
		double best_value = -1;
		for (int i = 1; i < sourceSentence.length+1; i++) {
		    double temp = params.getT(destSentence[j], sourceSentence[i-1])*params.getA(i, j+1, sourceSentence.length, destSentence.length);
		    if (temp > best_value) {
			best_value = temp;
			best_i = i;
		    }
		}
		alignmentArray[j] = best_i;
		a.probability *= best_value;
	    }
	}
	assert a.probability <= 1.0;

	a.setAlignment(alignmentArray);
	return a;	    
    }
	
     /**
     * Finds the best alignment between two sentences according to model 2
     * Brown et al. equation 84
     @param the source sentence
     @param the dest sentence
     @param parameters, should have T,A
     @return the best alignment
    */
    public static Alignment viterbi(String[] sourceSentence, String[] destSentence, ParameterAccessor params) {
	int[] alignmentArray = new int[destSentence.length];
	Alignment a = new Alignment();
	a.probability = 1;
	for (int j = 0; j < destSentence.length; j++) {
	    int best_i = -1;
	    double best_value = -1;
	    for (int i = 1; i < sourceSentence.length+1; i++) {
		double temp = params.getT(destSentence[j], sourceSentence[i-1])*params.getA(i, j+1, sourceSentence.length, destSentence.length);
		if (temp > best_value) {
		    best_value = temp;
		    best_i = i;
		}
	    }
	    alignmentArray[j] = best_i;
	    a.probability *= best_value;
	}
	a.setAlignment(alignmentArray);
	return a;
    }

    /**
     * Finds the APPROXIMATE best alignment between two sentences according to model 3, given a pegged position
     * Starts with an input alignment, and hill climbs from there. The probability attribute of the input alignment
     * must be set
     * Brown et al. equation 44; see: b^\infty
     @param the starting alignment--should generally be the result of model 2 viterbi
     @param the source sentence
     @param the dest sentence
     @param parameters, should have T,D,N,P
     @param pegged source position
     @param pegged dest position
     @return the approx best alignment
    */
    public static Alignment model3ViterbiApprox(Alignment alignment, String[] sourceSentence, String[] destSentence, Model3Params params, int peggedSource, int peggedDest) {
	if (alignment.probability == 0) { return alignment; }
	Alignment bestAlignment = alignment;
	
	Alignment[] neighbors = findNeighbors(alignment, sourceSentence.length);
	for (int i=0; i < neighbors.length; i++) {
	    if (peggedSource == neighbors[i].getAlignment()[peggedDest]) {
		neighbors[i].probability = 1;
		//this 'if' handles the case of a swap (as opposed to a simple move)
		if (neighbors[i].changedPosition2 != -1) {
		    
		    neighbors[i].probability *= neighborProbability(neighbors[i], alignment, neighbors[i].changedPosition2, sourceSentence, destSentence, params);
		    neighbors[i].probability /= alignment.probability;
		}
		assert !Double.isNaN(neighbors[i].probability);
		double x = neighborProbability(neighbors[i], alignment, neighbors[i].changedPosition1, sourceSentence, destSentence, params);
		neighbors[i].probability *= x;
		if ((neighbors[i].probability > 1.0) || (Double.isNaN(neighbors[i].probability))) {
		    neighbors[i].probability = 0;
		}
		assert (neighbors[i].probability <= 1.0);
		if (neighbors[i].probability > bestAlignment.probability) {
		    bestAlignment = new Alignment(neighbors[i].getAlignment());
		    bestAlignment.probability = neighbors[i].probability;
		}
	    }
	}
		
	if ((alignment == bestAlignment) || (bestAlignment.probability - alignment.probability < .0001)) {
	    return bestAlignment;
	}
	else {
	    return model3ViterbiApprox(bestAlignment, sourceSentence, destSentence, params, peggedSource, peggedDest);
	}
    }

    /**
     * Finds the APPROXIMATE best alignment between two sentences according to model 3
     * Starts with an input alignment, and hill climbs from there. The probability attribute of the input alignment
     * must be set
     * Brown et al. equation 44; see: b^\infty
     @param the starting alignment--should generally be the result of model 2 viterbi
     @param the source sentence
     @param the dest sentence
     @param parameters, should have T,D,N,P
     @return the approx best alignment
    */
    public static Alignment model3ViterbiApprox(Alignment alignment, String[] sourceSentence, String[] destSentence, Model3Params params) {
	if (alignment.probability == 0.0) return alignment;
	assert alignment.probability != 0.0;
	Alignment bestAlignment = alignment;
	
	Alignment[] neighbors = findNeighbors(alignment, sourceSentence.length);
	for (int i=0; i < neighbors.length; i++) {
	    neighbors[i].probability = 1;
	    if (neighbors[i].changedPosition2 != -1) {
		neighbors[i].probability *= neighborProbability(neighbors[i], alignment, neighbors[i].changedPosition2, sourceSentence, destSentence, params);
		neighbors[i].probability /= alignment.probability;
	    }
	    
	    neighbors[i].probability *= neighborProbability(neighbors[i], alignment, neighbors[i].changedPosition1, sourceSentence, destSentence, params);
	    //if (neighbors[i].probability >= 1.0) {System.out.println("here"); System.out.println(neighbors[i].probability); System.out.println(neighbors[i]); System.out.println(Arrays.toString(sourceSentence)); assert 1 ==3;}
	    if ((neighbors[i].probability > 1.0) || (Double.isNaN(neighbors[i].probability))) {
		    neighbors[i].probability = 0;
		}
		assert neighbors[i].probability <= 1.0;

	    if (neighbors[i].probability > bestAlignment.probability) {
		bestAlignment = new Alignment(neighbors[i].getAlignment());
		bestAlignment.probability = neighbors[i].probability;
	    }
	}
	if ((alignment == bestAlignment) || (bestAlignment.probability - alignment.probability < .0001)) {
	    return bestAlignment;
	}
	else {
	    return model3ViterbiApprox(bestAlignment, sourceSentence, destSentence, params);
	}
    }

    /**
     * Given an alignment and the length of the alignment's source sentence, finds all
     * neighbor alignments within a move or a swap of the alignment
     * Brown et al. equation 44--the stylized N function
     @param the starting alignment
     @param the length of the source sentence
     @return the neighborhood of alignments, in an array
    */
    public static Alignment[] findNeighbors(Alignment alignment, int sourceLength) {
	ArrayList<Alignment> neighbors = new ArrayList<Alignment>();
	int[] alignmentArray = alignment.getAlignment();
	int zeroCount = alignment.getFertility(0);
	boolean atZeroLimit;
	if (zeroCount+1 > alignment.size()/2) {
	    atZeroLimit = true;
	}
	else {
	    atZeroLimit = false;
	}
	//generates alignments within a move
	for (int i=0; i < alignment.size(); i++) {
	        
	    for (int j=0; j < sourceLength+1; j++) {
		if ((!(alignmentArray[i]==j)) && (j>0 || !atZeroLimit)){
		    int[] neighborArray = alignmentArray.clone();
		    neighborArray[i]=j;
		    Alignment a = new Alignment(neighborArray);
		    a.changedPosition1 = i;
		    neighbors.add(a);
		}
	    }
	}
	//generates alignments within a swap
	for (int i=0; i < alignment.size()-1; i++) {
	        
	    for (int j=i+1; j < alignment.size(); j++) {
		if (!(alignmentArray[i]==alignmentArray[j])) {
		    int[] neighborArray = alignmentArray.clone();
		    int temp = neighborArray[i];
		    neighborArray[i] = neighborArray[j];
		    neighborArray[j] = temp;
		    Alignment a = new Alignment(neighborArray);
		    a.changedPosition1 = j;
		    a.changedPosition2 = i;
		    neighbors.add(a);
		}
	    }
	}
	Alignment[] returnArray = new Alignment[neighbors.size()];	    
	neighbors.toArray(returnArray);
	return returnArray;
    }

    /**
     * Finds P(a,f|e)--that is, the probability of having a certain alignment and generated sentence, given
     * a fixed source sentence
     * See Kevin Knight's SMT workbook, section 25
     @param the alignment
     @param the dest sentence
     @param the source sentence
     @return the probability--P(a,f|e)
    */
    public static double getAlignmentAndTranslationProb(Alignment a, String[] destSentence, String[] sourceSentence, Model3Params params) {
	int m = destSentence.length;
	int l = sourceSentence.length;

	double product = 1.0;
	//Multiplies in the T value (translation) probabilities
	for (int j = 0; j < m; j++) {
	    product *= params.getModel3T(destSentence[j], a.getSourceWord(j, sourceSentence));
	}
	//Multiplies in the N value (fertility) probabilities
	for (int i = 1; i < l+1; i++) {
	    product *= params.getModel3N(a.getFertility(i), sourceSentence[i-1]);
	}
	//Multiplies in the D value (distortion) probabilities
	for (int j = 0; j < m; j++) {
	    int sourcePosition = a.getAlignment()[j];
	    if (sourcePosition != 0) {
		product *= params.getD(j+1, sourcePosition, l, m);
	    }
	}
	//Multiplies in values for spurious words (words generated from the null position)
	int nullFertility = a.getFertility(0);
	for (int i = 0; i < l; i++) {
	    product *= MathUtils.factorial(a.getFertility(i)) * Math.pow(MathUtils.factorial(nullFertility), -1) * MathUtils.binomialCoefficient(m-nullFertility, nullFertility) 
		* Math.pow(params.getP(), nullFertility) * Math.pow(1-params.getP(), m-2*nullFertility);
	}
	return product;
    }   
}